import importlib
import os
import subprocess
import sys
import warnings
from textwrap import dedent

import fsspec
import numpy
import numpy as np
import pytest

import tiffslide
from tiffslide import TiffFileError
from tiffslide import TiffSlide


@pytest.fixture
def slide(wsi_file):
    yield TiffSlide(wsi_file)


def test_image_detect_format(wsi_file):
    fmt = TiffSlide.detect_format(wsi_file)
    assert fmt is not None
    assert isinstance(fmt, str)


def test_image_open_incorrect_path():
    with pytest.raises(FileNotFoundError):
        TiffSlide("i-do-not-exist")


def test_image_open_unsupported_image(tmp_path):
    f = tmp_path.joinpath("test_file.svs")
    f.write_text("123")
    with pytest.raises(TiffFileError):
        TiffSlide(f)


def test_image_open(wsi_file):
    TiffSlide(wsi_file)


def test_image_context_manager(slide):
    with slide as t:
        # trigger full context
        _ = t.ts_zarr_grp


def test_image_repr(wsi_file, slide):
    assert os.path.basename(os.fspath(wsi_file)) in repr(slide)


def test_image_dimensions(slide):
    assert isinstance(slide.dimensions, tuple)
    assert len(slide.dimensions) == 2
    assert slide.dimensions[0] > 0 and slide.dimensions[1] > 0


def test_image_level_count(slide):
    assert slide.level_count >= 1


def test_image_level_dimensions(slide):
    assert isinstance(slide.level_dimensions, tuple)
    assert len(slide.level_dimensions) >= 1
    assert slide.dimensions == slide.level_dimensions[0]
    x0, y0 = slide.dimensions
    for x1, y1 in slide.level_dimensions[1:]:
        assert x1 < x0 and y1 < y0
        x0, y0 = x1, y1


def test_image_level_downsamples(slide):
    assert isinstance(slide.level_downsamples, tuple)
    assert len(slide.level_downsamples) >= 1
    for ds in slide.level_downsamples:
        assert isinstance(ds, float)
        assert ds >= 1.0


def test_image_properties(slide, svs_small_props):
    if slide.ts_tifffile.filename == "CMU-1-Small-Region.svs":
        assert slide.properties == svs_small_props
    else:
        pytest.skip("needs examples")


def test_image_get_best_level_for_downsample(slide):
    assert slide.get_best_level_for_downsample(1.0) == 0

    for lvl, ds in enumerate(slide.level_downsamples):
        # query exact downsample
        lvl_new = slide.get_best_level_for_downsample(ds)
        assert lvl == lvl_new

        # query smaller
        ds_smaller = numpy.nextafter(ds, 0)
        lvl_new = slide.get_best_level_for_downsample(ds_smaller)
        assert max(lvl - 1, 0) == lvl_new

        # query bigger
        ds_bigger = numpy.nextafter(ds, float("inf"))
        lvl_new = slide.get_best_level_for_downsample(ds_bigger)
        assert lvl == lvl_new


def test_image_read_region(slide):
    assert slide.read_region((0, 0), 0, (2220, 2967)).size == (2220, 2967)


def test_image_read_region_as_array(slide):
    assert slide.read_region((0, 0), 0, (2220, 2967), as_array=True).shape[:2] == (
        2967,
        2220,
    )


@pytest.mark.parametrize("use_embedded", [True, False])
def test_image_get_thumbnail(slide, use_embedded):
    thumb = slide.get_thumbnail((200, 200), use_embedded=use_embedded)
    assert max(thumb.size) == 200


def test_image_associated_images(slide):
    assoc = slide.associated_images
    assert repr(assoc)
    assert len(assoc) >= 0
    if slide.ts_tifffile.filename.endswith(".svs"):
        assert "thumbnail" in assoc
        assert "label" in assoc
        assert "macro" in assoc
        for key in assoc:
            img = assoc[key]
            assert img.size


def test_tiffslide_from_fsspec_buffer(wsi_file_urlpath):
    with fsspec.open(wsi_file_urlpath) as f:
        slide = TiffSlide(f)
        _ = slide.get_thumbnail((200, 200))


def test_tiffslide_from_fsspec_openfile(wsi_file_urlpath):
    of = fsspec.open(wsi_file_urlpath)
    slide = TiffSlide(of)
    _ = slide.get_thumbnail((200, 200))


def test_tiffslide_from_fsspec_urlpath(wsi_file_urlpath):
    slide = TiffSlide(wsi_file_urlpath)
    _ = slide.get_thumbnail((200, 200))


def test_tiffslide_reject_unsupported_file():
    with pytest.raises(ValueError):
        TiffSlide(dict())  # type: ignore


# === test aliases and fallbacks ========================================


def test_compat_open_slide(wsi_file):
    assert isinstance(tiffslide.open_slide(wsi_file), TiffSlide)


@pytest.mark.parametrize(
    "exc_name",
    [
        "OpenSlideUnsupportedFormatError",
        "OpenSlideError",
    ],
)
def test_compat_alias_exception(exc_name):
    with pytest.warns(UserWarning):
        cls = getattr(importlib.import_module("tiffslide"), exc_name)
    assert cls is TiffFileError


@pytest.mark.parametrize(
    "cls_name",
    [
        "OpenSlide",
        "ImageSlide",
    ],
)
def test_compat_alias_tiffslide(cls_name):
    with pytest.warns(UserWarning):
        cls = getattr(importlib.import_module("tiffslide"), cls_name)
    assert issubclass(cls, TiffSlide)


def test_compat_unsupported_abstractslide():
    with pytest.warns(UserWarning):
        with pytest.raises(ImportError):
            from tiffslide import AbstractSlide

            _ = AbstractSlide


def test_thread_safety(wsi_file, tmp_path):
    """reproduce threading issue reported in:

    https://github.com/bayer-science-for-a-better-life/tiffslide/issues/14

    """
    bug_py = tmp_path.joinpath("threading_bug.py")
    bug_py.write_text(
        dedent(
            f"""\
            import tiffslide
            import random
            from multiprocessing.pool import ThreadPool

            # number of threads
            num_threads = 8

            ts = tiffslide.TiffSlide({os.fspath(wsi_file)!r})
            w, h = ts.dimensions

            def read_region(slide):
                x = random.randint(0, max(0, w - 10))
                y = random.randint(0, max(0, h - 10))
                _ = slide.read_region((x, y), 0, (10, 10))

            with ThreadPool(num_threads) as pool:
                pool.starmap(read_region, [(ts, ) for _ in range(num_threads*2)])
            """
        )
    )
    for _ in range(10):
        out = subprocess.run([sys.executable, bug_py], capture_output=True)
        assert out.returncode == 0, out.stderr.decode()


def test_non_tiff_fallback(jpg_file):
    # noinspection PyProtectedMember
    from tiffslide import open_slide

    ts = open_slide(jpg_file)
    assert ts.properties
    assert ts.get_thumbnail((10, 10))
    assert ts.associated_images == {}
    assert ts.read_region((0, 0), 0, (10, 10))


@pytest.mark.parametrize(
    "loc,lvl",
    [
        pytest.param((-100, -100), 0, id="TL-level0"),
        pytest.param((0, -100), 0, id="TC-level0"),
        pytest.param((None, -100), 0, id="TR-level0"),
        pytest.param((None, 0), 0, id="MR-level0"),
        pytest.param((None, None), 0, id="BR-level0"),
        pytest.param((0, None), 0, id="BC-level0"),
        pytest.param((-100, None), 0, id="BL-level0"),
        pytest.param((-100, 0), 0, id="ML-level0"),
        pytest.param((0, 0), -1, id="negative-level"),
        pytest.param((0, 0), None, id="too-big-level"),
    ],
)
def test_read_region_padding_fully_oob(slide, loc, lvl):
    if loc[0] is None:
        loc = slide.dimensions[0], loc[1]
    if loc[1] is None:
        loc = loc[0], slide.dimensions[1]
    if lvl is None:
        lvl = slide.level_count

    size = (50, 50)

    def _all_padding(x):
        return np.all(x == 0)

    with pytest.warns(UserWarning, match="out-of-bounds"):
        r = slide.read_region(loc, lvl, size, as_array=True, padding=True)
    assert _all_padding(r), f"{loc!r}, {lvl}"


@pytest.mark.parametrize(
    "loc",
    [
        pytest.param((-100, -100), id="TL"),
        pytest.param((0, -100), id="TC"),
        pytest.param((None, -100), id="TR"),
        pytest.param((None, 0), id="MR"),
        pytest.param((None, None), id="BR"),
        pytest.param((0, None), id="BC"),
        pytest.param((-100, None), id="BL"),
        pytest.param((-100, 0), id="ML"),
    ],
)
def test_read_region_padding_partially_oob(slide, loc):
    size = (150, 150)

    if loc[0] is None:
        loc = slide.dimensions[0] - 50, loc[1]
    if loc[1] is None:
        loc = loc[0], slide.dimensions[1] - 50

    def _all_padding(x):
        return np.all(x == 0)

    with warnings.catch_warnings():
        warnings.simplefilter("error", UserWarning)
        r = slide.read_region(loc, 0, size, as_array=True, padding=True)

    assert not _all_padding(r), f"{loc!r}"


def test_read_region_nopadding_oob(slide):
    r = slide.read_region((-50, -50), 0, (100, 100), as_array=True, padding=False)
    assert r.shape[:2] == (50, 50)
    assert np.any(r != 0)

    locationbr = slide.dimensions[0] - 50, slide.dimensions[1] - 50
    r = slide.read_region(locationbr, 0, (100, 100), as_array=True, padding=False)
    assert r.shape[:2] == (50, 50)
    assert np.any(r != 0)

    r = slide.read_region((-100, -100), 0, (10, 10), as_array=True, padding=False)
    assert r.size == 0


def test_padding_no_padding(slide):
    lvl = slide.level_count - 1
    isize = width, height = slide.level_dimensions[lvl]
    rsize = width + 20, height + 40

    assert rsize == slide.read_region((-10, -10), lvl, rsize, padding=True).size
    assert isize == slide.read_region((-10, -10), lvl, rsize, padding=False).size


def test_padding_br_on_nonzero_level(slide):
    if slide.level_count == 1:
        pytest.skip("only one level")

    size = slide.level_dimensions[0]
    ds_lvl1 = slide.level_downsamples[1]
    loc = size[0] - int(10 * ds_lvl1), size[1] - int(10 * ds_lvl1)
    arr = slide.read_region(loc, 1, (20, 20), as_array=True)

    assert arr.ndim == 3
    assert arr.shape[0] == 20
    assert arr.shape[1] == 20
    assert arr.shape[2] in (1, 3)
    padding_bottom = arr[10:, :]
    padding_right = arr[:, 10:]
    assert np.all(padding_bottom == 0)
    assert np.all(padding_right == 0)


def test_no_numpy_scalar_overflow(slide):
    loc = np.array([255, 255], dtype=np.uint8)
    size = np.array([255, 255], dtype=np.uint8)

    with warnings.catch_warnings():
        warnings.simplefilter("error", RuntimeWarning)

        slide.read_region(loc, 0, size, as_array=True)


@pytest.mark.parametrize(
    "loc,lvl,size",
    [
        ((219, 219), 0, (48, 48)),
        ((219, 219), 1, (12, 12)),
        ((219, 219), 2, (3, 3)),
    ],
    ids=["lvl0", "lvl1", "lvl2"],
)
def test_read_region_intermediate_level_non_zero_loc(
    small_multilevel_img, loc, lvl, size
):
    # this test will fail, because of the way small_multilevel_img was constructed
    # the sizes of the base level were chosen so that the 4x lower zoom intermediate
    # levels will cause errors when reading from the intermediate levels, where
    # the location translated to the base level could be scaled asymmetrically.
    # see: https://github.com/bayer-science-for-a-better-life/tiffslide/issues/51
    slide = TiffSlide(small_multilevel_img)

    # return a sample of the checkerboard slide (we only need one channel)
    tile = slide.read_region(loc, lvl, size, as_array=True)[:, :, 0]

    # ensure we select an area that tests the checkerboard (in case refactoring breaks the test assumptions)
    assert not np.all(tile == tile[0, 0]), "test tile does not sample the checkerboard"

    # if the resulting tile is not symmetric it means we downsampled asymmetrically to the target level
    assert np.all(tile == tile.T), "tile does not sample the checkerboard symmetrically"



"""small helper script to compare the smallest zoom between openslide and tiffslide"""

from pathlib import Path
from pprint import pprint

import numpy as np
from openslide import OpenSlide
from PIL import Image

from tiffslide import TiffSlide


def write_lvl(fn):
    ts_slide = TiffSlide(fn)
    os_slide = OpenSlide(fn)

    print("ts ds:", ts_slide.level_downsamples)
    print("os ds:", os_slide.level_downsamples)

    c = ts_slide.properties.get("tiffslide.series-composition")
    if c:
        pprint(c)

    ts_lvl = ts_slide.level_count - 1
    os_lvl = os_slide.level_dimensions.index(ts_slide.level_dimensions[ts_lvl])
    size = ts_slide.level_dimensions[ts_lvl]

    ts_img = ts_slide.read_region((0, 0), ts_lvl, size)
    os_img = os_slide.read_region((0, 0), os_lvl, size)

    _p = Path(fn)
    ts_fn = f"{_p.stem}-{_p.suffix}_lvl{ts_lvl}_ts.png"
    os_fn = f"{_p.stem}-{_p.suffix}_lvl{ts_lvl}_os.png"
    err_fn = f"{_p.stem}-{_p.suffix}_lvl{ts_lvl}_err.png"

    ts_arr = np.array(ts_img)
    os_arr = np.array(os_img)
    os_arr = os_arr[:, :, :3]

    arr = np.abs(ts_arr.astype(int) - os_arr.astype(int))
    out = np.zeros_like(ts_arr, dtype=int)
    out = np.array(
        (out + (arr[:, :, :] * np.array([32, 32, 32]).reshape((1, 1, 3)))),
        dtype=np.uint8,
    )
    print("max difference:", np.max(arr))

    Image.fromarray(out).save(err_fn)

    ts_img.save(ts_fn)
    Image.fromarray(os_arr).save(os_fn)


if __name__ == "__main__":
    import sys

    if len(sys.argv) != 2:
        print("provide single filename")
        raise SystemExit(1)

    print(f"storing {sys.argv[1]}")
    write_lvl(sys.argv[1])
    print("done")



from __future__ import annotations

import enum
import hashlib
import os
import pathlib
import shutil
import sys
import urllib.request
from itertools import cycle
from itertools import groupby
from itertools import islice
from operator import attrgetter

import fsspec
import numpy as np
import pytest
import tifffile
from imagecodecs import imwrite

# openslide aperio test images
IMAGES_BASE_URL = "http://openslide.cs.cmu.edu/download/openslide-testdata/Aperio/"

try:
    APERIO_JP2000_RGB = tifffile.COMPRESSION.APERIO_JP2000_RGB
except AttributeError:
    # python3.7
    APERIO_JP2000_RGB = tifffile.TIFF.COMPRESSION.APERIO_JP2000_RGB


def md5(fn):
    if sys.version_info >= (3, 9):
        m = hashlib.md5(usedforsecurity=False)
    else:
        m = hashlib.md5()  # nosec B324
    with open(fn, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            m.update(chunk)
    return m.hexdigest()


class TestImageType(enum.Enum):
    DOWNLOAD_SMALLEST_CMU = enum.auto()
    GENERATE_PYRAMIDAL_IMG = enum.auto()
    GENERATE_PYRAMIDAL_1CH_16B_SVS = enum.auto()


def _wsi_files():
    """gather images"""

    local_test_images = os.environ.get("TIFFSLIDE_TEST_IMAGES", None)
    if local_test_images:
        _paths = []
        for img in pathlib.Path(local_test_images).glob("**/*"):
            # todo:
            #  - .zip files need to be tested
            #  - .zvi files need to be handled
            if img.suffix in {".svs", ".ndpi", ".scn", ".bif", ".tiff"}:
                _paths.append(img.absolute())

        def roundrobin(iterables):
            """roundrobin('ABC', 'D', 'EF') --> A D E B F C"""
            num_active = len(iterables)
            nexts = cycle(iter(it).__next__ for it in iterables)
            while num_active:
                try:
                    for next in nexts:
                        yield next()
                except StopIteration:
                    # Remove the iterator we just exhausted from the cycle.
                    num_active -= 1
                    nexts = cycle(islice(nexts, num_active))

        get_suffix = attrgetter("suffix")
        paths = [
            pytest.param(img, id=img.name)
            for img in roundrobin(
                [
                    sorted(list(group), key=lambda x: int(x.stat().st_size))
                    for key, group in groupby(
                        sorted(_paths, key=get_suffix), key=get_suffix
                    )
                ]
            )
        ]

    else:
        paths = [
            pytest.param(
                TestImageType.DOWNLOAD_SMALLEST_CMU, id="CMU-1-Small-Region.svs"
            ),
            pytest.param(
                TestImageType.GENERATE_PYRAMIDAL_IMG, id="generated-pyramidal"
            ),
            pytest.param(
                TestImageType.GENERATE_PYRAMIDAL_1CH_16B_SVS,
                id="generated-pyramidal-1ch-16b-svs",
            ),
        ]
    return paths


def _write_test_tiff(
    pth: os.PathLike[str],
    size: tuple[int, int],
    tile_size: int = 128,
    mpp: float = 0.5,
) -> None:
    def _num_pyramids(ldim: int, tsize: int) -> int:
        assert ldim > 0 and tsize > 0
        n = 0
        while ldim > tsize:
            ldim //= 2
            n += 1
        return n

    data = np.random.randint(0, 255, (size[0], size[1], 3), dtype=np.uint8)

    with tifffile.TiffWriter(pth, bigtiff=True, ome=True) as tif:
        im_height, im_width, _ = data.shape
        options0 = {}
        metadata = {}
        if mpp:
            metadata = {
                "PhysicalSizeX": mpp,
                "PhysicalSizeXUnit": "µm",
                "PhysicalSizeY": mpp,
                "PhysicalSizeYUnit": "µm",
            }
            options0["resolution"] = (1.0 / mpp, 1.0 / mpp, "MICROMETER")
        options = dict(
            tile=(tile_size, tile_size),
            photometric="rgb",
            compression="jpeg",
            metadata=metadata,
        )
        num_pyramids = _num_pyramids(max(im_height, im_width), tile_size)
        tif.write(
            data,
            subifds=num_pyramids,
            **options0,
            **options,
        )
        lvl_data = data
        for _ in range(num_pyramids):
            lvl_data = lvl_data[::2, ::2, :]
            tif.write(
                lvl_data,
                subfiletype=1,
                **options,
            )


def _write_test_svs_with_axes_YX_dtype_uint16(pth):
    """write a special tiff in svs format with single channel levels and dtype uint16

    author: One-Sixth https://github.com/One-sixth
    reported: https://github.com/bayer-science-for-a-better-life/tiffslide/issues/46
    """

    # fake data
    def gen_im(size_hw):
        while True:
            im = np.full(size_hw, 255, np.uint16)
            yield im

    thumbnail_im = np.zeros([762, 762], dtype=np.uint16)
    label_im = np.zeros([762, 762], dtype=np.uint16)
    macro_im = np.zeros([762, 762], dtype=np.uint16)

    # fake descriptions
    svs_desc = "Aperio Image Library Fake\nABC |AppMag = {mag}|Filename = {filename}|MPP = {mpp}"
    label_desc = "Aperio Image Library Fake\nlabel {W}x{H}"
    macro_desc = "Aperio Image Library Fake\nmacro {W}x{H}"

    tile_hw = (512, 512)
    # multi resolution
    multi_hw = [(10240, 10240), (5120, 5120), (2560, 2560)]
    mpp = 0.25
    mag = 40
    filename = "ASD"
    resolution_kw = {
        "resolution": (10000 / mpp, 10000 / mpp),
        "resolutionunit": "CENTIMETER",
    }

    # write to svs format
    with tifffile.TiffWriter(pth, bigtiff=True) as tif:
        kwargs = {
            "subifds": 0,
            "photometric": "MINISBLACK",
            "compression": APERIO_JP2000_RGB,
            "dtype": np.uint16,
            "metadata": None,
        }

        # write level 0
        tif.write(
            data=gen_im(tile_hw),
            shape=(*multi_hw[0], 1),
            tile=tile_hw[::-1],
            description=svs_desc.format(mag=mag, filename=filename, mpp=mpp),
            **resolution_kw,
            **kwargs,
        )
        # write thumbnail image
        tif.write(data=thumbnail_im, description="", **kwargs)

        # write level 1 to N
        for hw in multi_hw[1:]:
            tif.write(
                data=gen_im(tile_hw),
                shape=(*hw, 1),
                tile=tile_hw[::-1],
                description="",
                **resolution_kw,
                **kwargs,
            )

        # write label
        tif.write(
            data=label_im,
            subfiletype=1,
            description=label_desc.format(W=label_im.shape[1], H=label_im.shape[0]),
            **kwargs,
        )
        # write marco
        tif.write(
            data=macro_im,
            subfiletype=9,
            description=macro_desc.format(W=macro_im.shape[1], H=macro_im.shape[0]),
            **kwargs,
        )


def _write_test_jpg(
    pth: os.PathLike[str],
    size: tuple[int, int],
):
    arr = np.random.randint(0, 255, size=(size[0], size[1], 3), dtype=np.uint8)
    imwrite(pth, arr)


@pytest.fixture(scope="session")
def jpg_file(tmp_path_factory):
    jpg_pth = tmp_path_factory.mktemp("test_images").joinpath("test.jpg")
    _write_test_jpg(jpg_pth, (1024, 1024))
    yield jpg_pth


@pytest.fixture(scope="session", params=_wsi_files())
def wsi_file(request, tmp_path_factory):
    """download the smallest aperio test image svs or use local"""
    if request.param == TestImageType.DOWNLOAD_SMALLEST_CMU:
        small_image = "CMU-1-Small-Region.svs"
        small_image_md5 = "1ad6e35c9d17e4d85fb7e3143b328efe"
        data_dir = pathlib.Path(__file__).parent / "data"

        data_dir.mkdir(parents=True, exist_ok=True)
        img_fn = data_dir / small_image

        if not img_fn.is_file():
            # download svs from openslide test images
            url = IMAGES_BASE_URL + small_image
            with urllib.request.urlopen(url) as response, open(  # nosec B310
                img_fn, "wb"
            ) as out_file:
                shutil.copyfileobj(response, out_file)

        if md5(img_fn) != small_image_md5:  # pragma: no cover
            shutil.rmtree(img_fn)
            pytest.fail("incorrect md5")

    elif request.param == TestImageType.GENERATE_PYRAMIDAL_IMG:
        img_fn = tmp_path_factory.mktemp("_generated_test_tiffs").joinpath(
            "_small_pyramid.tiff"
        )
        _write_test_tiff(img_fn, (4096, 4096))

    elif request.param == TestImageType.GENERATE_PYRAMIDAL_1CH_16B_SVS:
        img_fn = tmp_path_factory.mktemp("_generated_test_tiffs").joinpath(
            "_small_pyramid_2.svs"
        )
        _write_test_svs_with_axes_YX_dtype_uint16(img_fn)

    else:
        img_fn = request.param

    yield img_fn.absolute()


@pytest.fixture()
def small_multilevel_img(tmp_path):
    # small_multilevel_img is a 3 level 48x48 tile-size image
    # with a size that causes the smaller level sizes to have
    # slightly different downsamples when calculated from sizes
    # alone.
    import tifffile

    pth = tmp_path.joinpath("small_multilevel.tiff")

    size = (480, 370, 3)
    tile_size = 48
    data = np.ones(size, dtype="uint8")

    for iy, y in enumerate(range(0, size[0], tile_size)):
        for ix, x in enumerate(range(0, size[1], tile_size)):
            val = (iy + ix) % 2
            data[y : y + tile_size, x : x + tile_size] = 100 * (val + 1)

    # (46000, 32914), (11500, 8228)

    with tifffile.TiffWriter(pth, bigtiff=False, ome=True) as tif:
        im_height, im_width, _ = size
        options0 = {}
        metadata = {
            "PhysicalSizeX": 0.5,
            "PhysicalSizeXUnit": "µm",
            "PhysicalSizeY": 0.5,
            "PhysicalSizeYUnit": "µm",
        }
        options0["resolution"] = (2.0, 2.0, "MICROMETER")
        options = dict(
            tile=(tile_size, tile_size),
            photometric="rgb",
            metadata=metadata,
        )
        if tuple(map(int, tifffile.__version__.split("."))) > (2022, 5, 4):
            options["compression"] = "jpeg"
            options["compressionargs"] = {"level": 100}
        else:
            options["compression"] = "jpeg", 100

        tif.write(
            data,
            subifds=2,
            **options0,
            **options,
        )
        lvl_data = data
        for _ in range(2):
            lvl_data = lvl_data[::4, ::4, :]
            tif.write(
                lvl_data,
                subfiletype=1,
                **options,
            )

    yield pth


@pytest.fixture
def wsi_file_urlpath(wsi_file):
    if wsi_file.stat().st_size > 100 * 1024 * 1024:
        pytest.skip("reduce ram usage of tests")
    urlpath = f"memory://{wsi_file.name}"
    fs: fsspec.AbstractFileSystem = fsspec.get_filesystem_class("memory")()
    of = fsspec.open(urlpath, mode="wb")
    with of as f:
        f.write(wsi_file.read_bytes())
    try:
        yield urlpath
    finally:
        fs.rm(wsi_file.name)


@pytest.fixture(scope="session")
def svs_small_props():
    yield {
        'aperio.AppMag': 20,
        'aperio.Date': '12/29/09',
        'aperio.Filename': 'CMU-1',
        'aperio.Filtered': 5,
        'aperio.Focus Offset': 0.0,
        'aperio.Header': 'Aperio Image Library v11.2.1 \r\n'
                         '46000x32914 [42673,5576 2220x2967] (240x240) JPEG/RGB Q=30'
                         ';Aperio Image Library v10.0.51\r\n'
                         '46920x33014 [0,100 46000x32914] (256x256) JPEG/RGB Q=30',
        'aperio.ImageID': 1004486,
        'aperio.Left': 25.691574,
        'aperio.LineAreaXOffset': 0.019265,
        'aperio.LineAreaYOffset': -0.000313,
        'aperio.LineCameraSkew': -0.000424,
        'aperio.MPP': 0.499,
        'aperio.OriginalHeight': 32914,
        'aperio.OriginalWidth': 46000,
        'aperio.Originalheight': 33014,
        'aperio.Parmset': 'USM Filter',
        'aperio.ScanScope ID': 'CPAPERIOCS',
        'aperio.StripeWidth': 2040,
        'aperio.Time': '09:59:15',
        'aperio.Top': 23.449873,
        'aperio.User': 'b414003d-95c6-48b0-9369-8010ed517ba7',
        'tiff.ImageDescription': 'Aperio Image Library v11.2.1 \r\n'
                                 '46000x32914 [42673,5576 2220x2967] (240x240) '
                                 'JPEG/RGB Q=30;Aperio Image Library v10.0.51\r\n'
                                 '46920x33014 [0,100 46000x32914] (256x256) JPEG/RGB '
                                 'Q=30|AppMag = 20|StripeWidth = '
                                 '2040|ScanScope ID = CPAPERIOCS|Filename = '
                                 'CMU-1|Date = 12/29/09|Time = 09:59:15|User = '
                                 'b414003d-95c6-48b0-9369-8010ed517ba7|Parmset = USM '
                                 'Filter|MPP = 0.4990|Left = 25.691574|Top = '
                                 '23.449873|LineCameraSkew = '
                                 '-0.000424|LineAreaXOffset = '
                                 '0.019265|LineAreaYOffset = -0.000313|Focus Offset = '
                                 '0.000000|ImageID = 1004486|OriginalWidth = '
                                 '46920|Originalheight = 33014|Filtered = '
                                 '5|OriginalWidth = 46000|OriginalHeight = 32914',
        'tiffslide.series-index': 0,
        'tiffslide.series-axes': 'YXS',
        'tiffslide.background-color': None,
        'tiffslide.bounds-height': None,
        'tiffslide.bounds-width': None,
        'tiffslide.bounds-x': None,
        'tiffslide.bounds-y': None,
        'tiffslide.comment': 'Aperio Image Library v11.2.1 \r\n'
                             '46000x32914 [42673,5576 2220x2967] (240x240) JPEG/RGB '
                             'Q=30;Aperio Image Library v10.0.51\r\n'
                             '46920x33014 [0,100 46000x32914] (256x256) JPEG/RGB '
                             'Q=30|AppMag = 20|StripeWidth = 2040|ScanScope ID = '
                             'CPAPERIOCS|Filename = CMU-1|Date = 12/29/09|Time = '
                             '09:59:15|User = '
                             'b414003d-95c6-48b0-9369-8010ed517ba7|Parmset = USM '
                             'Filter|MPP = 0.4990|Left = 25.691574|Top = '
                             '23.449873|LineCameraSkew = -0.000424|LineAreaXOffset = '
                             '0.019265|LineAreaYOffset = -0.000313|Focus Offset = '
                             '0.000000|ImageID = 1004486|OriginalWidth = '
                             '46920|Originalheight = 33014|Filtered = 5|OriginalWidth '
                             '= 46000|OriginalHeight = 32914',
        'tiffslide.level[0].downsample': 1.0,
        'tiffslide.level[0].height': 2967,
        'tiffslide.level[0].tile-height': 240,
        'tiffslide.level[0].tile-width': 240,
        'tiffslide.level[0].width': 2220,
        'tiffslide.mpp-x': 0.499,
        'tiffslide.mpp-y': 0.499,
        'tiffslide.objective-power': 20,
        'tiffslide.quickhash-1': None,
        'tiffslide.vendor': 'aperio',
    }  # fmt: skip



import sys
from hashlib import sha1

from flask import Flask
from flask import abort
from flask import make_response
from flask import render_template
from flask import url_for

from tiffslide.deepzoom import MinimalComputeAperioDZGenerator

app = Flask(__name__)

# note: this is not good practice, but it'll do for the sake of simplicity
_dzgen = None


if sys.version_info >= (3, 9):

    def hash_fn(x: str) -> str:
        return sha1(x.encode(), usedforsecurity=False).hexdigest()

else:

    def hash_fn(x: str) -> str:
        return sha1(x.encode()).hexdigest()  # nosec


@app.route("/")
def index():
    # noinspection PyProtectedMember
    fn = _dzgen._openfile.path  # just so we don't need to worry about caching
    su = url_for("dzi", image_fn=hash_fn(fn))
    return render_template("index.html", filename=fn, slide_url=su)


@app.route("/<image_fn>.dzi")
def dzi(image_fn):
    resp = make_response(_dzgen.get_dzi())
    resp.mimetype = "application/xml"
    return resp


@app.route("/<image_fn>_files/<int:level>/<int:col>_<int:row>.jpeg")
def tile(image_fn, level, col, row):
    try:
        tile_bytes = _dzgen.get_tile(level, col, row)
    except (KeyError, IndexError):
        return abort(404)

    resp = make_response(tile_bytes)
    resp.mimetype = "image/jpeg"
    return resp


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--port", type=int, default=8000, help="server port")
    parser.add_argument("--bind", type=str, default="127.0.0.1", help="server address")
    parser.add_argument("svs_file", help="svs file")
    args = parser.parse_args()

    _dzgen = MinimalComputeAperioDZGenerator(args.svs_file)

    app.run(host=args.bind, port=args.port)



import contextlib
import itertools
import os
import warnings
from pathlib import Path

import numpy as np
import pytest

pytestmark = pytest.mark.compat

OPENSLIDE_TESTDATA_DIR = os.getenv("OPENSLIDE_TESTDATA_DIR", None)
_FILES = {
    "svs": [
        "Aperio/CMU-1-JP2K-33005.svs",
        "Aperio/CMU-1-Small-Region.svs",
        "Aperio/CMU-1.svs",
        "Aperio/CMU-2.svs",
        "Aperio/CMU-3.svs",
        "Aperio/JP2K-33003-1.svs",
        "Aperio/JP2K-33003-2.svs",
    ],
    "generic": [
        "Generic-TIFF/CMU-1.tiff",
    ],
    "hamamatsu": [
        "Hamamatsu/CMU-1.ndpi",
        "Hamamatsu/CMU-2.ndpi",
        "Hamamatsu/CMU-3.ndpi",
        "Hamamatsu/OS-1.ndpi",
        "Hamamatsu/OS-2.ndpi",
        "Hamamatsu/OS-3.ndpi",
    ],
    "leica": [
        "Leica/Leica-1.scn",
        "Leica/Leica-2.scn",
        # todo: these two can't be opened by OpenSlide
        # "Leica/Leica-3.scn",
        # "Leica/Leica-Fluorescence-1.scn",
    ],
    "ventana": [
        "Ventana/OS-1.bif",
        "Ventana/OS-2.bif",
    ],
}
FILES = {}

if OPENSLIDE_TESTDATA_DIR is not None:
    for key, fns in _FILES.items():
        for fn in fns:
            k = f"{key}-{fn.split('/')[1]}"
            FILES[k] = os.path.join(OPENSLIDE_TESTDATA_DIR, fn)


MODULES = ["tiffslide", "openslide"]


def matches(fn, vendor=None, filename=None, ext=None):
    if vendor is not None:
        return any(Path(x).name == Path(fn).name for x in _FILES[vendor])
    elif filename is not None:
        return fn == file_name
    elif ext is not None:
        return Path(fn).suffix == f".{ext}"
    else:
        raise ValueError("all none")


@pytest.fixture(params=list(FILES))
def file_name(request):
    f = FILES[request.param]
    if not os.path.isfile(f):
        pytest.xfail("missing local test file")
    else:
        yield f


@pytest.fixture()
def ts_slide(file_name):
    from tiffslide import TiffSlide

    yield TiffSlide(file_name)


@pytest.fixture()
def os_slide(file_name):
    from openslide import OpenSlide

    yield OpenSlide(file_name)


def test_openslide_testdata_dir_env():
    assert os.getenv("OPENSLIDE_TESTDATA_DIR") is not None


def test_dimensions(ts_slide, os_slide):
    assert ts_slide.dimensions == os_slide.dimensions


def test_level_count(ts_slide, os_slide, file_name):
    if matches(file_name, vendor="hamamatsu"):
        pytest.xfail("'ndpi' no computed levels")

    assert ts_slide.level_count == os_slide.level_count


def test_level_dimensions(ts_slide, os_slide, file_name):
    if matches(file_name, vendor="hamamatsu"):
        pytest.xfail("'ndpi' no computed levels")

    assert ts_slide.level_dimensions == os_slide.level_dimensions


def test_level_downsamples(ts_slide, os_slide, file_name):
    if matches(file_name, vendor="hamamatsu"):
        pytest.xfail("'ndpi' no computed levels")

    assert ts_slide.level_downsamples == os_slide.level_downsamples


def test_strict_subset_level_dimensions(ts_slide, os_slide):
    # test that all available tiffslide levels are in os_slide
    assert set(ts_slide.level_dimensions).issubset(os_slide.level_dimensions)


def test_strict_subset_level_downsamples(ts_slide, os_slide):
    # test that all available tiffslide downsamples are in os_slide
    assert set(ts_slide.level_downsamples).issubset(os_slide.level_downsamples)


def test_read_region_equality_level_min(ts_slide, os_slide, file_name):
    exact = True
    if "JP2K-33003" in file_name:
        warnings.warn(
            f"JP2K file {file_name} is tested to be almost equal (not exactly equal)!",
            stacklevel=2,
        )
        exact = False

    width, height = ts_slide.dimensions

    ws = range(0, width, width // 5)
    hs = range(0, height, height // 5)
    for loc in itertools.product(ws[:-1], hs[:-1]):
        ts_img = ts_slide.read_region(loc, 0, (128, 128))
        os_img = os_slide.read_region(loc, 0, (128, 128))

        ts_arr = np.array(ts_img)
        os_arr = np.array(os_img)
        # np.testing.assert_equal(os_arr[:, :, 3], 255)
        os_arr = os_arr[:, :, :3]

        if exact:
            np.testing.assert_equal(ts_arr, os_arr)
        else:
            np.testing.assert_allclose(ts_arr, os_arr, atol=1, rtol=0)


@contextlib.contextmanager
def _debug_image_differences(arr0, arr1, name=None):
    try:
        yield
    except BaseException:
        from PIL import Image

        pth = os.getenv("DEBUG_OPENSLIDE_IMAGE_DIR", None)
        if pth is not None:
            Image.fromarray(arr0).save(os.path.join(pth, f"{name}-0-tiffslide.png"))
            Image.fromarray(arr1).save(os.path.join(pth, f"{name}-1-openslide.png"))

        raise


def test_read_region_equality_level_intermediate(ts_slide, os_slide, file_name):
    if ts_slide.level_count <= 2:
        pytest.skip("requires at least 3 levels")

    exact = True
    if "JP2K-33003" in file_name:
        warnings.warn(
            f"JP2K file {file_name} is tested to be almost equal (not exactly equal)!",
            stacklevel=2,
        )
        exact = False

    level = ts_slide.level_count // 2
    width, height = ts_slide.dimensions

    ws = range(0, width, width // 5)
    hs = range(0, height, height // 5)
    for loc in itertools.product(ws[:-1], hs[:-1]):
        ts_img = ts_slide.read_region(loc, level, (256, 256))
        os_img = os_slide.read_region(loc, level, (256, 256))

        ts_arr = np.array(ts_img)
        os_arr = np.array(os_img)
        # np.testing.assert_equal(os_arr[:, :, 3], 255)
        os_arr = os_arr[:, :, :3]

        with _debug_image_differences(
            ts_arr,
            os_arr,
            f"{os.path.basename(file_name)}-x{loc[0]}y{loc[1]}-lvl{level}",
        ):
            if exact:
                np.testing.assert_equal(ts_arr, os_arr)
            else:
                np.testing.assert_allclose(ts_arr, os_arr, atol=1, rtol=0)


def test_read_region_equality_level_common_max(ts_slide, os_slide, file_name):
    exact = True
    if "JP2K-33003" in file_name:
        warnings.warn(
            f"JP2K file {file_name} is tested to be almost equal (not exactly equal)!",
            stacklevel=2,
        )
        exact = False

    ts_lvl = ts_slide.level_count - 1
    os_lvl = os_slide.level_dimensions.index(ts_slide.level_dimensions[ts_lvl])
    size = ts_slide.level_dimensions[ts_lvl]

    if size[0] >= 5000 or size[1] >= 5000:
        pytest.skip("smallest common level is too big...")

    ts_img = ts_slide.read_region((0, 0), ts_lvl, size)
    os_img = os_slide.read_region((0, 0), os_lvl, size)

    ts_arr = np.array(ts_img)
    os_arr = np.array(os_img)
    os_arr = os_arr[:, :, :3]

    max_difference = np.max(np.abs(ts_arr.astype(int) - os_arr.astype(int)))
    assert max_difference <= (0 if exact else 1)



"""tiffslide._kerchunk

Test implementation of serialization and deserialization code to and from kerchunk
Don't rely on this API until it's public.

"""
from __future__ import annotations

import json
import os
import sys
from collections import ChainMap
from io import StringIO
from typing import TYPE_CHECKING
from typing import Any
from typing import TypedDict

import fsspec
from imagecodecs.numcodecs import register_codecs

from tiffslide.tiffslide import TiffSlide

if TYPE_CHECKING:
    from fsspec.implementations.reference import ReferenceFileSystem


__all__ = [
    "to_kerchunk",
    "from_kerchunk",
]


TIFFSLIDE_SPEC_VERSION = 1
KERCHUNK_SPEC_VERSION = 1


class KerchunkSpec(TypedDict):
    """simple kerchunk version 1 spec"""

    version: int
    templates: dict[str, str]
    gen: list[Any]
    refs: dict[str, Any]


def to_kerchunk(
    slide: TiffSlide,
    *,
    urlpath: str,
    templatename: str | None = None,
) -> KerchunkSpec:
    """take a tiffslide instance and extract a kerchunk representation"""
    if slide.ts_tifffile.filename is None:
        raise ValueError("can't kerchunk a slide that's not backed by a named file")

    kw = {}
    if templatename is not None:
        kw["templatename"] = templatename

    if not isinstance(urlpath, str):
        urlpath = os.fspath(urlpath)
    if urlpath.endswith(slide.ts_tifffile.filename):
        urlpath = urlpath[: -len(slide.ts_tifffile.filename)]

    combined = KerchunkSpec(
        version=KERCHUNK_SPEC_VERSION,
        templates={},
        gen=[],
        refs={},
    )
    for idx, series in enumerate(slide.ts_tifffile.series):
        with StringIO() as f, series.aszarr() as store:
            try:
                store.write_fsspec(
                    f,
                    urlpath,
                    groupname=f"s{idx}",
                    version=KERCHUNK_SPEC_VERSION,
                    **kw,
                )
            except ValueError as err:
                if "incomplete chunks" in str(err):
                    continue
                raise
            kc = json.loads(f.getvalue())

        # combine individual series
        combined["templates"].update(kc["templates"])
        combined["gen"].extend(kc["gen"])

        combined_refs = combined["refs"]
        refs = kc["refs"]

        # insert .zgroup
        _zgroup = refs.pop(".zgroup")
        if ".zgroup" in combined_refs:
            assert combined_refs[".zgroup"] == _zgroup
        combined_refs[".zgroup"] = _zgroup

        # insert .zattrs
        if ".zattrs" not in combined_refs:
            combined_refs[".zattrs"] = json.dumps(
                {
                    "tiffslide.spec_version": TIFFSLIDE_SPEC_VERSION,
                    "tiffslide.properties": slide.properties,
                }
            )
        assert ".zattrs" not in refs

        # insert refs
        combined_refs.update(refs)

    return combined


def from_kerchunk(
    kc: KerchunkSpec,
    *,
    urlpath: str | None = None,
    storage_options: dict[str, Any] | None = None,
) -> TiffSlide:
    """deserialize a TiffSlide from a kerchunk specification"""
    if urlpath:
        # replace urlpath in kerchunk spec
        templates = kc.get("templates", {}).copy()
        if len(set(templates.values())) != 1:
            raise NotImplementedError(
                "urlpath replacement support only for length 1 templates"
            )
        templates = {key: urlpath for key in templates}
        kc = ChainMap({"templates": templates}, kc)  # type: ignore

    if storage_options is None:
        storage_options = {}

    fs: ReferenceFileSystem = fsspec.filesystem(
        "reference",
        fo=kc,
        **storage_options,
    )
    zattrs = json.loads(fs.cat_file(".zattrs"))

    if "tiffslide.spec_version" not in zattrs or "tiffslide.properties" not in zattrs:
        raise ValueError("")

    if zattrs["tiffslide.spec_version"] != 1:
        raise NotImplementedError("TiffSlide spec version unsupported")
    properties = zattrs["tiffslide.properties"]

    # register codecs now...
    register_codecs(verbose=False)

    # prepare the TiffSlide instance
    inst = object.__new__(TiffSlide)
    inst.__dict__["properties"] = properties
    inst._tifffile = fs  # fixme: ...
    return inst


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser("tiffslide._kerchunk")
    parser.add_argument("urlpath", help="fsspec compatible urlpath to image")
    parser.add_argument(
        "--storage-options",
        type=json.loads,
        help="json encoded storage options",
        default=None,
    )
    args = parser.parse_args()

    # open slide, serialize and reopen
    _slide = TiffSlide(args.urlpath, storage_options=args.storage_options)
    k = to_kerchunk(_slide, urlpath=args.urlpath)
    # kc_slide = from_kerchunk(k)
    print(json.dumps(k, separators=(",", ":")))



"""tiffslide

a somewhat drop-in replacement for openslide-python using tifffile and zarr

"""
from warnings import warn

from tiffslide._types import PathOrFileOrBufferLike
from tiffslide.tiffslide import PROPERTY_NAME_BACKGROUND_COLOR
from tiffslide.tiffslide import PROPERTY_NAME_BOUNDS_HEIGHT
from tiffslide.tiffslide import PROPERTY_NAME_BOUNDS_WIDTH
from tiffslide.tiffslide import PROPERTY_NAME_BOUNDS_X
from tiffslide.tiffslide import PROPERTY_NAME_BOUNDS_Y
from tiffslide.tiffslide import PROPERTY_NAME_COMMENT
from tiffslide.tiffslide import PROPERTY_NAME_MPP_X
from tiffslide.tiffslide import PROPERTY_NAME_MPP_Y
from tiffslide.tiffslide import PROPERTY_NAME_OBJECTIVE_POWER
from tiffslide.tiffslide import PROPERTY_NAME_QUICKHASH1
from tiffslide.tiffslide import PROPERTY_NAME_VENDOR
from tiffslide.tiffslide import NotTiffSlide
from tiffslide.tiffslide import TiffFileError
from tiffslide.tiffslide import TiffSlide

try:
    from tiffslide._version import version as __version__
except ImportError:  # pragma: no cover
    __version__ = "not-installed"

__all__ = ["TiffSlide", "TiffFileError"]


def __getattr__(name):  # type: ignore
    """support some drop-in behavior"""
    # alias the most important bits
    if name in {"OpenSlideUnsupportedFormatError", "OpenSlideError"}:
        warn(
            f"compatibility: aliasing tiffslide.TiffFileError to {name!r}", stacklevel=2
        )
        return TiffFileError
    elif name == "OpenSlide":
        warn(f"compatibility: aliasing tiffslide.TiffSlide to {name!r}", stacklevel=2)
        return TiffSlide
    elif name == "ImageSlide":
        warn(
            f"compatibility: aliasing tiffslide.NotTiffSlide to {name!r}", stacklevel=2
        )
        return NotTiffSlide
    # warn if internals are imported that we don't support
    if name in {"AbstractSlide", "__library_version__"}:
        warn(f"{name!r} is not provided by tiffslide", stacklevel=2)
    raise AttributeError(name)


def open_slide(filename: PathOrFileOrBufferLike) -> TiffSlide:
    """drop-in helper function

    Note: this will always try to fall back to the NotTiffSlide class
      in case creating a TiffSlide instance fails. If this is undesired
      use `tiffslide.TiffSlide` directly instead of `tiffslide.open_slide`
    """
    try:
        return TiffSlide(filename)
    except TiffFileError as original_err:
        try:
            return NotTiffSlide(filename)
        except Exception as fallback_err:
            # I want this fallback to raise the original error
            # so that user code can catch TiffFileError.
            # this is bending the __cause__ rules a bit...
            raise original_err from fallback_err



"""helpers for broken wsi"""
from __future__ import annotations

import functools
import inspect
import os
import os.path
import warnings
from typing import Any

import tifffile


def fix_description_tag_encoding(
    filename: str | os.PathLike[str],
    *,
    errors: str = "ignore",
    no_dry_run: bool = False,
) -> int:
    """overwrites broken description tags"""
    filename = os.path.expanduser(filename)
    tf = tifffile.TiffFile(filename, mode="r+")

    applied_fixes = False
    for page in tf.pages:
        try:
            description_tag = page.tags["ImageDescription"]
        except KeyError:
            continue
        value = description_tag.value
        if isinstance(value, bytes) and value:
            if not no_dry_run:
                print("# would fix tag", description_tag)
                print("- ", repr(value)[2:-1])
                print("+ ", repr(value.decode(errors=errors).encode())[2:-1])
            else:
                print("# fixing tag", description_tag)
                new_value = value.decode("ascii", errors=errors)
                if len(new_value.encode()) > len(value):
                    raise RuntimeError("fixed string is longer than original?")
                description_tag.overwrite(new_value)
            applied_fixes = True
    if not applied_fixes:
        print("# found no problems with description tag encoding")
    elif applied_fixes and not no_dry_run:
        print("\n# run the command again with `--no-dry-run` to apply fixes")
        print("# PLEASE BACKUP THE ORIGINAL FILE BEFORE!")
    return 0


def monkey_patch_description_tag_encoding() -> None:
    from tifffile.tifffile import TiffTags

    original_valueof_method = TiffTags.valueof
    if getattr(original_valueof_method, "__monkey_patched__", False):
        return

    sig = inspect.signature(original_valueof_method)
    if "key" not in sig.parameters:
        raise AssertionError("monkey patch won't work on this tifffile version")

    @functools.wraps(original_valueof_method)
    def patched_valueof(*args: Any, **kwargs: Any) -> Any:
        ba = sig.bind(*args, **kwargs)
        ba.apply_defaults()
        value = original_valueof_method(*ba.args, **ba.kwargs)
        if ba.arguments["key"] == 270 and isinstance(value, bytes):
            value = value.decode(errors="ignore")
        return value

    patched_valueof.__monkey_patched__ = True  # type: ignore

    warnings.warn(
        "Monkey patching `tifffile.tifffile.TiffTags` to recover non-ascii description tags.\n"
        "IT IS NOT RECOMMENDED TO USE THIS WORKAROUND to load files that contain non-ascii\n"
        "characters in their image description tags. To fix the WSI File, please run:\n"
        "$ python -m tiffslide.repair fix-description-tag-encoding <wsi-file>\n",
        RuntimeWarning,
        stacklevel=2,
    )

    # monkey patch the TiffTags class
    TiffTags.valueof = patched_valueof


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        description="Collection of helper scripts for broken wsi files"
    )
    subparsers = parser.add_subparsers(required=True)

    parser0 = subparsers.add_parser("fix-description-tag-encoding")
    parser0.add_argument("filename", help="local filename for wsi")
    parser0.add_argument("--no-dry-run", action="store_true", help="modify wsi")
    parser0.set_defaults(func=fix_description_tag_encoding)

    args = parser.parse_args()
    kw = vars(args).copy()
    del kw["func"]

    raise SystemExit(args.func(**kw))



"""tiffslide._compat

compatibility layer to support loading non-tiff images

"""
from __future__ import annotations

import json
import os.path
import sys
from pathlib import PurePath
from types import MappingProxyType
from types import TracebackType
from typing import TYPE_CHECKING
from typing import Any
from typing import Literal
from typing import Mapping
from typing import Sequence

import numpy as np
import zarr
from imagecodecs import __version__ as _imagecodecs_version
from imagecodecs import imread

try:
    from tiffslide._version import version as _tiffslide_version
except ImportError:  # pragma: no cover
    _tiffslide_version = "not-installed"

if TYPE_CHECKING:
    from numpy.typing import NDArray

__all__ = [
    "NotTiffFile",
]


class NotTiffFile:
    def __init__(
        self,
        arg: Any,
        mode: Literal["rb"] | None = None,
        name: str | None = None,
        *_args: Any,
        **_kwargs: Any,
    ) -> None:
        if mode is not None and mode != "rb":
            raise ValueError("mode must be 'rb'")
        if name is None:
            if isinstance(arg, str):
                name = os.path.basename(arg)
            elif isinstance(arg, PurePath):
                name = arg.name
            elif hasattr(arg, "fullname"):
                name = arg.fullname
        self.filename = name

        array, codec = imread(arg, return_codec=True)

        self.pages = [NotTiffPage(array, codec=codec.__name__)]
        self.series = [NotTiffPageSeries(self.pages)]

    def close(self) -> None:
        pass

    def __enter__(self) -> NotTiffFile:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        return None

    def __getattr__(self, item: str) -> Any:
        if item.startswith("is_"):
            return False
        raise AttributeError(item)


class NotTiffPageSeries:
    axes = "YXS"  # todo: check if it's always this

    def __init__(self, pages: NotTiffPages) -> None:
        self.pages = pages

    def __getitem__(self, item: int) -> NotTiffPage:
        return self.pages[item]

    def __len__(self) -> int:
        return len(self.pages)

    @property
    def ndim(self) -> int:
        return self.pages[0].ndim

    @property
    def shape(self) -> tuple[int, ...]:
        return self.pages[0].shape

    @property
    def levels(self) -> list[NotTiffPageSeries]:
        return [self]

    @property
    def size(self) -> int:
        return self.pages[0].size

    def aszarr(self, **_: Any) -> zarr.storage.MemoryStore:
        return self.pages[0].aszarr()

    def asarray(self) -> NDArray[np.uint8]:
        return self.pages[0].asarray()


NotTiffPages = Sequence["NotTiffPage"]


class NotTiffPage:
    tags: Mapping[Any, Any] = MappingProxyType({})
    tilelength = 0
    tilewidth = 0

    def __init__(self, array: NDArray[np.uint8], codec: str):
        self._array = array
        self._codec = codec

    @property
    def description(self) -> str:
        data = json.dumps(
            {
                "tiffslide.__version__": _tiffslide_version,
                "imagecodecs.__version__": _imagecodecs_version,
                "codec": self._codec,
            },
            separators=(",", ":"),
        )
        return f"tiffslide={data}"

    @property
    def ndim(self) -> int:
        return self._array.ndim

    @property
    def shape(self) -> tuple[int, ...]:
        return self._array.shape

    @property
    def size(self) -> int:
        return self._array.size

    def aszarr(self) -> zarr.storage.MemoryStore:
        return zarr.creation.array(self._array).store

    def asarray(self) -> NDArray[np.uint8]:
        return self._array



from tifffile import ZarrStore

from tiffslide import TiffSlide

# noinspection PyProtectedMember
from tiffslide._zarr import get_zarr_chunk_sizes


def test_get_chunk_sizes(wsi_file):
    chunk_sizes = get_zarr_chunk_sizes(TiffSlide(wsi_file).zarr_group)
    assert chunk_sizes.ndim in (2, 3)
    assert chunk_sizes.sum() > 0


def test_decode_only_once(wsi_file, monkeypatch):
    ts = TiffSlide(wsi_file)

    called_keys = []
    _ZarrStore___getitem__ = ZarrStore.__getitem__

    def _cnt_getitem(self, key):
        nonlocal called_keys
        called_keys.append(key)
        return _ZarrStore___getitem__(self, key)

    monkeypatch.setattr(ZarrStore, "__getitem__", _cnt_getitem)

    _ = ts.read_region((0, 0), 0, (1, 1), as_array=True)
    assert len(called_keys) == len(set(called_keys))



import importlib
import os
import platform
import random
import subprocess
import sys
import warnings
from getpass import getpass
from itertools import product

import numpy as np
import pytest

OPENSLIDE_TESTDATA_DIR = os.getenv("OPENSLIDE_TESTDATA_DIR", None)
FILES = {
    "svs": "Aperio/CMU-2.svs",
    "svs_jp2k": "Aperio-JP2K/TCGA-05-4395-01Z-00-DX1.20205276-ca16-46b2-914a-fe5e576a5cf9.svs",
    "generic": "Generic-TIFF/CMU-1.tiff",
    "hamamatsu": "Hamamatsu/OS-3.ndpi",
    "leica": "Leica/Leica-2.scn",
    "ventana": "Ventana/OS-2.bif",
}
if OPENSLIDE_TESTDATA_DIR is None:
    pytestmark = pytest.mark.skip

else:
    for key, fn in FILES.items():
        FILES[key] = os.path.join(OPENSLIDE_TESTDATA_DIR, fn)


MODULES = ["tiffslide", "openslide"]


@pytest.fixture(scope="module")
def root_pw(request):
    capture = request.config.pluginmanager.getplugin("capturemanager")
    capture.suspend_global_capture(in_=True)
    if not sys.stdin.isatty() and sys.stdout.isatty():
        warnings.warn("DISK CACHE INVALIDATION NEEDS ROOT, but running non-interactive")
        root_pw = None
    else:
        root_pw = getpass(
            "\n\nPLEASE ENTER ROOT PW (empty disables disk cache invalidation): "
        )
    capture.resume_global_capture()
    yield root_pw


def try_to_clear_disk_cache(root_pass):
    env = os.environ.copy()
    if root_pw:  # type: ignore
        env["SUDO_PASSWORD"] = root_pass

    system = platform.system()
    if system == "Linux":
        # from: https://github.com/sdvillal/jagged/blob/6fc83aa11e/jagged/benchmarks/utils.py#L95-L118
        max_size = "1000G"
        drop_level = 3

        if 0 != os.system(  # nosec B605
            f'vmtouch -e -f -q -m {max_size} "{OPENSLIDE_TESTDATA_DIR}"'
        ):
            subprocess.run(
                r"""printf '%%s\n' "$SUDO_PASSWORD" """
                r"""| sudo -p "" -S -- sh -c 'sync && echo %d > /proc/sys/vm/drop_caches'"""
                % drop_level,
                env=env,
                check=True,
                shell=True,  # nosec B602
            )

    elif system == "Darwin":
        subprocess.run(
            r"""printf '%s\n' "$SUDO_PASSWORD" | sudo -p "" -S -- sh -c 'sync && purge'""",
            env=env,
            check=True,
            shell=True,
        )

    elif system == "Windows":
        warnings.warn("not implemented for windows yet")


@pytest.fixture(
    params=[pytest.param((ft, m), id=f"{m}-{ft}") for ft, m in product(FILES, MODULES)]
)
def slide_with_tile_size(request, root_pw):
    """yield a slide together with the internal tile size"""
    file_type, module_name = request.param
    file_name = FILES[file_type]

    # get the corresponding module
    mod = importlib.import_module(module_name)
    open_slide = getattr(mod, "open_slide")

    try_to_clear_disk_cache(root_pw)

    slide = open_slide(file_name)

    # attach tile_size
    try:
        t_size = (
            int(slide.properties[f"{module_name}.level[0].tile-width"]),
            int(slide.properties[f"{module_name}.level[0].tile-height"]),
        )
    except KeyError:
        warnings.warn("recovering tile_size via tiffslide")
        ts_cls = getattr(importlib.import_module("tiffslide"), "TiffSlide")
        p = ts_cls(file_name).properties
        t_size = (
            int(p[f"tiffslide.level[0].tile-width"]),
            int(p[f"tiffslide.level[0].tile-height"]),
        )

    try:
        yield slide, t_size
    finally:
        slide.close()


def get_locations(slide, order, tsize):
    def _locations(s, ts, c_order):
        w, h = s.dimensions
        tw, th = ts

        out = []
        if c_order:
            for y in range(0, h, th):
                for x in range(0, w, tw):
                    out.append((x, y))
        else:
            for x in range(0, w, tw):
                for y in range(0, h, th):
                    out.append((x, y))
        return out

    if order == "seqrow":
        locations = _locations(slide, tsize, c_order=True)
    elif order == "seqcol":
        locations = _locations(slide, tsize, c_order=False)
    elif order == "random":
        locations = _locations(slide, tsize, c_order=True)
        random.seed(42)
        random.shuffle(locations)
    else:
        raise ValueError(order)
    return locations


@pytest.mark.parametrize("order", ["seqrow", "seqcol", "random"])
def test_read_tiles_as_pil(order, slide_with_tile_size, benchmark):
    slide, tsize = slide_with_tile_size
    locations = get_locations(slide, order, tsize)
    lociter = iter(locations[len(locations) // 2 :])

    def setup():
        return (next(lociter), 0, tsize), {}

    read_tile = slide.read_region

    benchmark.pedantic(read_tile, setup=setup, rounds=64, iterations=1, warmup_rounds=1)


@pytest.mark.parametrize("order", ["seqrow", "seqcol", "random"])
def test_read_tiles_as_numpy(order, slide_with_tile_size, benchmark):
    slide, tsize = slide_with_tile_size
    locations = get_locations(slide, order, tsize)
    lociter = iter(locations[len(locations) // 2 :])

    def setup():
        return (next(lociter), 0, tsize), {}

    _read_tile = slide.read_region
    if slide.__class__.__name__ == "OpenSlide":

        def read_tile_np(loc, level, size):
            return np.array(_read_tile(loc, level, size))

    elif slide.__class__.__name__ == "TiffSlide":

        def read_tile_np(loc, level, size):
            return _read_tile(loc, level, size, as_array=True)

    else:
        raise NotImplementedError("unknown class")

    benchmark.pedantic(
        read_tile_np, setup=setup, rounds=64, iterations=1, warmup_rounds=1
    )



import os
import platform

import pytest

from tiffslide import TiffSlide
from tiffslide._kerchunk import from_kerchunk
from tiffslide._kerchunk import to_kerchunk


@pytest.fixture
def slide(wsi_file):
    yield TiffSlide(wsi_file)


def test_to_kerchunk(slide, wsi_file):
    kc = to_kerchunk(slide, urlpath=wsi_file)
    assert kc["version"] == 1
    assert kc["gen"] == []
    if platform.system() != "Windows":
        assert os.fspath(wsi_file) in set(kc["templates"].values())
    assert kc["refs"]


def test_from_kerchunk(slide, wsi_file):
    kc = to_kerchunk(slide, urlpath=wsi_file)

    ts = from_kerchunk(kc)
    assert ts.properties
    assert ts.read_region((0, 0), 0, (100, 100), as_array=True).size > 0


def test_kerchunked_repr(slide, wsi_file):
    kc = to_kerchunk(slide, urlpath=wsi_file)

    ts = from_kerchunk(kc)
    assert repr(ts)



"""small helper script to compare tiles between openslide and tiffslide"""

import itertools
from pathlib import Path

import numpy as np
from openslide import OpenSlide
from PIL import Image

from tiffslide import TiffSlide


def write_tiles(fn):
    ts_slide = TiffSlide(fn)
    os_slide = OpenSlide(fn)

    width, height = ts_slide.dimensions

    ws = range(0, width, width // 5)
    hs = range(0, height, height // 5)
    for loc in itertools.product(ws[:-1], hs[:-1]):
        ts_img = ts_slide.read_region(loc, 0, (512, 512))
        os_img = os_slide.read_region(loc, 0, (512, 512))

        _p = Path(fn)

        ts_fn = f"{_p.stem}-{_p.suffix}_x{loc[0]}y{loc[1]}_ts.png"
        os_fn = f"{_p.stem}-{_p.suffix}_x{loc[0]}y{loc[1]}_os.png"
        err_fn = f"{_p.stem}-{_p.suffix}_x{loc[0]}y{loc[1]}_zzz_err.png"

        ts_arr = np.array(ts_img)
        os_arr = np.array(os_img)
        # assert np.all(os_arr[:, :, 3] == 255)
        os_arr = os_arr[:, :, :3]

        arr = np.abs(ts_arr.astype(int) - os_arr.astype(int))
        out = np.zeros_like(ts_arr, dtype=int)
        out = np.array(
            (out + (arr[:, :, :] * np.array([32, 32, 32]).reshape((1, 1, 3)))),
            dtype=np.uint8,
        )
        print("max difference:", np.max(arr))

        Image.fromarray(out).save(err_fn)

        ts_img.save(ts_fn)
        Image.fromarray(os_arr).save(os_fn)


if __name__ == "__main__":
    import sys

    if len(sys.argv) != 2:
        print("provide single filename")
        raise SystemExit(1)

    print(f"tiling {sys.argv[1]}")
    write_tiles(sys.argv[1])
    print("done")



from __future__ import annotations

import math
import os.path
import sys
from collections import defaultdict
from fractions import Fraction
from functools import cached_property
from itertools import count
from types import TracebackType
from typing import TYPE_CHECKING
from typing import Any
from typing import Iterator
from typing import Literal
from typing import Mapping
from typing import Optional
from typing import TypeVar
from typing import overload
from warnings import warn
from xml.etree import ElementTree

import numpy as np
import tifffile
import zarr
from fsspec.core import url_to_fs
from fsspec.implementations.local import LocalFileSystem
from fsspec.implementations.reference import ReferenceFileSystem
from PIL import Image
from tifffile import TiffFile
from tifffile import TiffFileError as TiffFileError
from tifffile import TiffPage
from tifffile import TiffPageSeries
from tifffile.tifffile import svs_description_metadata

from tiffslide._compat import NotTiffFile
from tiffslide._types import OpenFileLike
from tiffslide._types import PathOrFileOrBufferLike
from tiffslide._types import SeriesCompositionInfo
from tiffslide._types import Slice3D
from tiffslide._types import TiffFileIO
from tiffslide._zarr import get_zarr_depth_and_dtype
from tiffslide._zarr import get_zarr_selection
from tiffslide._zarr import get_zarr_store

if TYPE_CHECKING:
    import numpy.typing as npt


__all__ = [
    "PROPERTY_NAME_COMMENT",
    "PROPERTY_NAME_VENDOR",
    "PROPERTY_NAME_QUICKHASH1",
    "PROPERTY_NAME_BACKGROUND_COLOR",
    "PROPERTY_NAME_OBJECTIVE_POWER",
    "PROPERTY_NAME_MPP_X",
    "PROPERTY_NAME_MPP_Y",
    "PROPERTY_NAME_BOUNDS_X",
    "PROPERTY_NAME_BOUNDS_Y",
    "PROPERTY_NAME_BOUNDS_WIDTH",
    "PROPERTY_NAME_BOUNDS_HEIGHT",
    "TiffSlide",
    "TiffFileError",
    "NotTiffSlide",
]

# all relevant tifffile version numbers work with this.
_TIFFFILE_VERSION = tuple(
    int(x) if x.isdigit() else x for x in tifffile.__version__.split(".")
)

# === Constants to support drop-in ===
PROPERTY_NAME_COMMENT = "tiffslide.comment"
PROPERTY_NAME_VENDOR = "tiffslide.vendor"
PROPERTY_NAME_QUICKHASH1 = "tiffslide.quickhash-1"
PROPERTY_NAME_BACKGROUND_COLOR = "tiffslide.background-color"
PROPERTY_NAME_OBJECTIVE_POWER = "tiffslide.objective-power"
PROPERTY_NAME_MPP_X = "tiffslide.mpp-x"
PROPERTY_NAME_MPP_Y = "tiffslide.mpp-y"
PROPERTY_NAME_BOUNDS_X = "tiffslide.bounds-x"
PROPERTY_NAME_BOUNDS_Y = "tiffslide.bounds-y"
PROPERTY_NAME_BOUNDS_WIDTH = "tiffslide.bounds-width"
PROPERTY_NAME_BOUNDS_HEIGHT = "tiffslide.bounds-height"

# prevent pillow>=9.1.0 deprecation warning
try:
    _ANTIALIAS = Image.Resampling.LANCZOS
except AttributeError:
    _ANTIALIAS = Image.ANTIALIAS
try:
    _NEAREST = Image.Resampling.NEAREST
except AttributeError:
    _NEAREST = Image.NEAREST


class TiffSlide:
    """
    tifffile backed whole slide image container emulating openslide.OpenSlide
    """

    def __init__(
        self,
        filename: PathOrFileOrBufferLike,
        *,
        tifffile_options: dict[str, Any] | None = None,
        storage_options: dict[str, Any] | None = None,
    ) -> None:
        """TiffSlide constructor

        Parameters
        ----------
        filename:
            a local filename or a fsspec urlpath or a file object
        tifffile_options:
            a dictionary with keyword arguments passed to the TiffFile constructor
        storage_options:
            a dictionary with keyword arguments passed to fsspec
        """
        # tifffile instance, can raise TiffFileError
        self._tifffile = _prepare_tifffile(
            filename,
            storage_options=storage_options,
            tifffile_options=tifffile_options,
            _cls=TiffFile,
        )

    @property
    def ts_tifffile(self) -> TiffFile:
        """get the underlying tifffile instance"""
        # backwards compatibility
        if isinstance(self._tifffile, ReferenceFileSystem):
            raise RuntimeError(
                "instance is backed by kerchunk: no ts_tifffile available"
            )
        elif not isinstance(self._tifffile, (TiffFile, NotTiffFile)):
            raise NotImplementedError(
                f"instance backed by {type(self._tifffile).__name__}"
            )
        return self._tifffile

    def __enter__(self) -> TiffSlide:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self.close()

    def close(self) -> None:
        try:
            grp = self.__dict__.pop("zarr_group")
        except KeyError:
            pass
        else:
            try:
                grp.close()
            except AttributeError:
                pass
            del grp
        self.ts_tifffile.close()

    def __repr__(self) -> str:
        fn = _get_filename(self._tifffile)
        r = repr(fn) if fn else "<unknown filename ...>"
        return f"{type(self).__name__}({r})"

    @classmethod
    def detect_format(
        cls,
        filename: PathOrFileOrBufferLike,
        *,
        tifffile_options: dict[str, Any] | None = None,
        storage_options: dict[str, Any] | None = None,
    ) -> str | None:
        """return the detected format as a str or None if unknown/unimplemented"""
        try:
            tf = _prepare_tifffile(
                filename,
                tifffile_options=tifffile_options,
                storage_options=storage_options,
                _cls=TiffFile,
            )
        except TiffFileError:
            return None
        with tf as t:
            return _PropertyParser.detect_format(t)

    @cached_property
    def dimensions(self) -> tuple[int, int]:
        """return the width and height of level 0"""
        prop = self.properties
        return (
            prop["tiffslide.level[0].width"],
            prop["tiffslide.level[0].height"],
        )

    @cached_property
    def level_count(self) -> int:
        """return the number of levels"""
        prop = self.properties
        lvl = 1
        while f"tiffslide.level[{lvl}].width" in prop:
            lvl += 1
        return lvl

    @cached_property
    def level_dimensions(self) -> tuple[tuple[int, int], ...]:
        """return the dimensions of levels as a list"""
        prop = self.properties
        lvl_dims = [self.dimensions]
        for lvl in count(1):
            try:
                lvl_dim = (
                    prop[f"tiffslide.level[{lvl}].width"],
                    prop[f"tiffslide.level[{lvl}].height"],
                )
            except KeyError:
                break
            else:
                lvl_dims.append(lvl_dim)
        return tuple(lvl_dims)

    @cached_property
    def level_downsamples(self) -> tuple[float, ...]:
        """return the downsampling factors of levels as a list"""
        w0, h0 = self.dimensions
        return tuple(((w0 / w) + (h0 / h)) / 2.0 for w, h in self.level_dimensions)

    @cached_property
    def properties(self) -> dict[str, Any]:
        """image properties / metadata as a dict"""
        return _PropertyParser(self._tifffile).parse()

    @cached_property
    def associated_images(self) -> _LazyAssociatedImagesDict:
        """return associated images as a mapping of names to PIL images"""
        idx = self.properties["tiffslide.series-index"]
        series = self.ts_tifffile.series[idx + 1 :]
        return _LazyAssociatedImagesDict(series)

    def get_best_level_for_downsample(self, downsample: float) -> int:
        """return the best level for a given downsampling factor"""
        if downsample <= 1.0:
            return 0
        for lvl, ds in enumerate(self.level_downsamples):
            if ds > downsample:
                return lvl - 1
        return self.level_count - 1

    @cached_property
    def zarr_group(self) -> zarr.hierarchy.Group:
        """return the tiff image as a zarr-like group

        NOTE: this is extra functionality and not part of the drop-in behaviour
        """
        try:
            _num_decode = os.environ["TIFFSLIDE_NUM_DECODE_THREADS"]
        except KeyError:
            num_decode_threads = None  # half of num CPU
        else:
            if _num_decode:
                num_decode_threads = int(_num_decode)
            else:
                num_decode_threads = None
        store = get_zarr_store(
            self.properties, self._tifffile, num_decode_threads=num_decode_threads
        )
        return zarr.open_group(store, mode="r")

    @property
    def ts_zarr_grp(self) -> zarr.hierarchy.Group:
        """use .zarr_group instead"""
        # backwards compatibility
        return self.zarr_group

    @overload
    def read_region(
        self,
        location: tuple[int, int],
        level: int,
        size: tuple[int, int],
    ) -> Image.Image:
        ...

    @overload
    def read_region(
        self,
        location: tuple[int, int],
        level: int,
        size: tuple[int, int],
        *,
        as_array: Literal[False] = ...,
        padding: bool = ...,
    ) -> Image.Image:
        ...

    @overload
    def read_region(
        self,
        location: tuple[int, int],
        level: int,
        size: tuple[int, int],
        *,
        as_array: Literal[True] = ...,
        padding: bool = ...,
    ) -> npt.NDArray[np.int_]:
        ...

    def read_region(
        self,
        location: tuple[int, int],
        level: int,
        size: tuple[int, int],
        *,
        as_array: bool = False,
        padding: bool = True,
    ) -> Image.Image | npt.NDArray[np.int_]:
        """return the requested region as a PIL.Image

        Parameters
        ----------
        location :
            pixel location (x, y) in level 0 of the image
        level :
            target level used to read the image
        size :
            size (width, height) of the requested region
        as_array :
            if True, return the region as numpy array
        padding :
            if True, will ensure that the size of the returned image is deterministic.
        """
        base_x, base_y = map(int, location)
        _rw, _rh = map(int, size)
        axes = self.properties["tiffslide.series-axes"]

        try:
            if level < 0:
                raise IndexError
            level_w, level_h = self.level_dimensions[level]
        except IndexError:
            if not padding:
                raise IndexError(f"level={level} out of range")
            warn(
                f"level={level} is out-of-bounds, but padding is requested",
                stacklevel=2,
            )

            depth, dtype = get_zarr_depth_and_dtype(self.zarr_group, axes)
            return np.zeros((_rh, _rw, depth), dtype=dtype)

        rx0, ry0 = self._read_region_loc_transform((base_x, base_y), level)
        rx1 = rx0 + _rw
        ry1 = ry0 + _rh

        in_bound = 0 <= rx0 and rx1 <= level_w and 0 <= ry0 and ry1 <= level_h
        requires_padding = padding and not in_bound

        if requires_padding:
            # compute padding
            pad_x0 = _clip(-rx0, 0, _rw)
            pad_x1 = _clip(rx1 - level_w, 0, _rw)
            pad_y0 = _clip(-ry0, 0, _rh)
            pad_y1 = _clip(ry1 - level_h, 0, _rh)

        if not in_bound:
            # crop coord to valid zone
            rx0 = _clip(rx0, 0, level_w)
            rx1 = _clip(rx1, 0, level_w)
            ry0 = _clip(ry0, 0, level_h)
            ry1 = _clip(ry1, 0, level_h)

        selection: Slice3D
        if axes == "YXS":
            selection = slice(ry0, ry1), slice(rx0, rx1), slice(None)
        elif axes == "CYX":
            selection = slice(None), slice(ry0, ry1), slice(rx0, rx1)
        elif axes == "YX":
            selection = slice(ry0, ry1), slice(rx0, rx1), ...
        else:
            raise NotImplementedError(f"axes={axes!r}")

        arr: npt.NDArray[np.int_] = get_zarr_selection(
            self.zarr_group,
            selection=selection,
            level=level,
        )

        if axes == "CYX":
            arr = arr.transpose((1, 2, 0))
        elif axes == "YX":
            arr = arr[..., np.newaxis]

        if requires_padding:
            if arr.shape[0] == 0 or arr.shape[1] == 0:
                warn(
                    f"location={location!r}, level={level}, size={size!r} is out-of-bounds, but padding is requested",
                    stacklevel=2,
                )

            # noinspection PyUnboundLocalVariable
            arr = np.pad(
                arr,
                ((pad_y0, pad_y1), (pad_x0, pad_x1), (0, 0)),
                mode="constant",
                constant_values=0,
            )

        if as_array:
            return arr
        elif axes == "YX":
            return Image.fromarray(arr[..., 0])
        else:
            return Image.fromarray(arr)

    def _read_region_loc_transform(
        self, location: tuple[int, int], level: int
    ) -> tuple[int, int]:
        """return the location at the provided level

        Notes
        -----
        Overwrite in subclasses in case you want to change the default
        interpretation of the `loc` argument in `read_region()`.

        """
        base_x, base_y = location
        level_ds = self.level_downsamples[level]
        rx0 = int(base_x / level_ds)
        ry0 = int(base_y / level_ds)
        return rx0, ry0

    def get_thumbnail(
        self, size: tuple[int, int], *, use_embedded: bool = False
    ) -> Image.Image:
        """return the thumbnail of the slide as a PIL.Image with a maximum size

        Parameters
        ----------
        size:
            width,height tuple defining maximum size of the thumbnail in each direction.
            the thumbnail itself keeps the image aspect ratio
        use_embedded:
            if True uses the embedded thumbnail in the image (if available and smaller
            than the highest level) to generate the thumbnail image

        """
        if (
            use_embedded
            and "thumbnail" in self.associated_images
            and size <= self.associated_images.series_map["thumbnail"].shape[1::-1]
        ):
            thumb_byte_size = self.associated_images.series_map["thumbnail"].size
        else:
            thumb_byte_size = -1

        slide_w, slide_h = self.dimensions
        thumb_w, thumb_h = size
        downsample = max(slide_w / thumb_w, slide_h / thumb_h)
        level = self.get_best_level_for_downsample(downsample)

        idx = self.properties["tiffslide.series-index"]
        level_byte_size = self.ts_tifffile.series[idx].levels[level].size

        if 0 < thumb_byte_size < level_byte_size:
            # read the embedded thumbnail if it uses fewer bytes
            img = self.associated_images["thumbnail"]
        else:
            # read the best suited level
            _level_dimensions = self.level_dimensions[level]
            img = self.read_region((0, 0), level, _level_dimensions)

        # now composite the thumbnail
        thumb = Image.new(
            mode=img.mode,
            size=img.size,
            color=f"#{self.properties[PROPERTY_NAME_BACKGROUND_COLOR] or 'ffffff'}",
        )
        thumb.paste(img, box=None, mask=None)
        try:
            thumb.thumbnail(size, _ANTIALIAS)
        except ValueError:
            # see: https://github.com/python-pillow/Pillow/blob/95cff6e959/src/libImaging/Resample.c#L559-L588
            thumb.thumbnail(size, _NEAREST)
        return thumb


class NotTiffSlide(TiffSlide):
    # noinspection PyMissingConstructor
    def __init__(
        self,
        filename: PathOrFileOrBufferLike,
        *,
        tifffile_options: dict[str, Any] | None = None,
        storage_options: dict[str, Any] | None = None,
    ) -> None:
        # tifffile instance, can raise TiffFileError
        self._tifffile = _prepare_tifffile(
            filename,
            storage_options=storage_options,
            tifffile_options=tifffile_options,
            _cls=NotTiffFile,
        )

    @classmethod
    def detect_format(
        cls,
        filename: PathOrFileOrBufferLike,
        *,
        tifffile_options: dict[str, Any] | None = None,
        storage_options: dict[str, Any] | None = None,
    ) -> str | None:
        """return the detected format as a str or None if unknown/unimplemented"""
        try:
            tf = _prepare_tifffile(
                filename,
                tifffile_options=tifffile_options,
                storage_options=storage_options,
                _cls=NotTiffFile,
            )
        except ValueError:
            return None
        with tf as t:
            # noinspection PyProtectedMember
            return t.pages[0]._codec


class _LazyAssociatedImagesDict(Mapping[str, Image.Image]):
    """lazily load associated images"""

    def __init__(self, series: list[TiffPageSeries]):
        self.series_map: dict[str, TiffPageSeries] = {s.name.lower(): s for s in series}
        self._m: dict[str, Image.Image] = {}

    def __repr__(self) -> str:
        args = ", ".join(
            f"{name!r}: <lazy-loaded PIL.Image.Image size={s.shape[1]}x{s.shape[0]} ...>"
            for name, s in self.series_map.items()
        )
        # pretend to be a normal dictionary
        return f"{{{args}}}"

    def __getitem__(self, k: str) -> Image.Image:
        if k in self._m:
            return self._m[k]
        else:
            s = self.series_map[k]
            self._m[k] = img = Image.fromarray(s.asarray())
            return img

    def __len__(self) -> int:
        return len(self.series_map)

    def __iter__(self) -> Iterator[str]:
        yield from self.series_map


TF = TypeVar("TF", TiffFile, NotTiffFile)


def _prepare_tifffile(
    fb: PathOrFileOrBufferLike,
    *,
    tifffile_options: dict[str, Any] | None = None,
    storage_options: dict[str, Any] | None = None,
    _cls: type[TF] = TiffFile,
) -> TF:
    """prepare a TiffFile instance

    Allows providing fsspec urlpaths as well as fsspec OpenFile instances directly.

    Parameters
    ----------
    fb:
        an urlpath like string, a fsspec OpenFile like instance or a buffer like instance
    tifffile_options:
        keyword arguments passed to tifffile.TiffFile
    storage_options:
        keyword arguments passed to fsspec AbstractFileSystem.open()
    """
    tf_kw: dict[str, Any] = tifffile_options or {}
    st_kw: dict[str, Any] = storage_options or {}

    def _warn_unused_storage_options(kw: Any) -> None:
        if kw:
            warn(
                "storage_options ignored when providing file or buffer like object",
                stacklevel=3,
            )

    _monkey_patch()

    if isinstance(fb, TiffFileIO):
        # provided an IO stream like instance
        _warn_unused_storage_options(st_kw)

        return _cls(fb, **tf_kw)

    elif isinstance(fb, OpenFileLike):
        # provided a fsspec compatible OpenFile instance
        _warn_unused_storage_options(st_kw)

        fs, path = fb.fs, fb.path

        # set name for tifffile.FileHandle
        if "name" not in tf_kw:
            if hasattr(fb, "full_name"):
                name = os.path.basename(fb.full_name)  # type: ignore
            else:
                name = os.path.basename(path)
            tf_kw["name"] = name

        return _cls(fs.open(path), **tf_kw)

    elif isinstance(fb, (str, os.PathLike)):
        # provided a string like url
        urlpath = os.fspath(fb)
        fs, path = url_to_fs(urlpath, **st_kw)
        if isinstance(fs, LocalFileSystem):
            return _cls(path, **tf_kw)
        else:
            # set name for tifffile.FileHandle
            if "name" not in tf_kw:
                tf_kw["name"] = os.path.basename(path)

            return _cls(fs.open(path), **tf_kw)

    else:
        # let's try anyways ...
        _warn_unused_storage_options(st_kw)

        return _cls(fb, **tf_kw)


# --- property / metadata related functionality -------------------------------


class _PropertyParser:
    """parse tiffslide properties for different slide types"""

    vendor_map = dict(
        svs="aperio",
        scn="leica",
        bif="ventana",
        ndpi="hamamatsu",
        philips="philips_tiff"
        # add more when needed
    )

    def __init__(self, tf: TiffFile) -> None:
        self._tf = tf

    @staticmethod
    def new_metadata() -> dict[str, Any]:
        return dict.fromkeys(
            [
                PROPERTY_NAME_COMMENT,
                PROPERTY_NAME_VENDOR,
                PROPERTY_NAME_QUICKHASH1,
                PROPERTY_NAME_BACKGROUND_COLOR,
                PROPERTY_NAME_OBJECTIVE_POWER,
                PROPERTY_NAME_MPP_X,
                PROPERTY_NAME_MPP_Y,
                PROPERTY_NAME_BOUNDS_X,
                PROPERTY_NAME_BOUNDS_Y,
                PROPERTY_NAME_BOUNDS_WIDTH,
                PROPERTY_NAME_BOUNDS_HEIGHT,
            ]
        )

    @classmethod
    def detect_format(cls, tf: TiffFile) -> str:
        for prop, vendor in cls.vendor_map.items():
            if getattr(tf, f"is_{prop}"):
                return vendor
        return "generic-tiff"

    @classmethod
    def collect_level_info(cls, series: TiffPageSeries) -> dict[str, Any]:
        # calculate level info
        md = {}
        if series.ndim not in (2, 3):
            raise NotImplementedError(
                "currently no support for series.ndim not in (2, 3)"
            )

        axes = md["tiffslide.series-axes"] = series.axes

        if axes == "YXS":
            h0, w0, _ = map(int, series.shape)
            level_dimensions = ((lvl.shape[1], lvl.shape[0]) for lvl in series.levels)
        elif axes == "CYX":
            _, h0, w0 = map(int, series.shape)
            level_dimensions = ((lvl.shape[2], lvl.shape[1]) for lvl in series.levels)
        elif axes == "YX":
            h0, w0 = map(int, series.shape)
            level_dimensions = ((lvl.shape[1], lvl.shape[0]) for lvl in series.levels)
        else:
            raise NotImplementedError(f"series with axes={axes!r} not supported yet")

        for lvl, (width, height) in enumerate(level_dimensions):
            downsample = ((w0 / width) + (h0 / height)) / 2.0
            page = series.levels[lvl][0]
            md[f"tiffslide.level[{lvl}].downsample"] = downsample
            md[f"tiffslide.level[{lvl}].height"] = int(height)
            md[f"tiffslide.level[{lvl}].width"] = int(width)
            md[f"tiffslide.level[{lvl}].tile-height"] = page.tilelength
            md[f"tiffslide.level[{lvl}].tile-width"] = page.tilewidth
        return md

    @classmethod
    def recover_mpp(cls, series: TiffPageSeries) -> dict[str, Any]:
        """recover mpp from tiff tags"""
        page0 = series[0]
        md: dict[str, Any] = {}

        try:
            resolution_unit = page0.tags["ResolutionUnit"].value
            x_resolution = Fraction(*page0.tags["XResolution"].value)
            y_resolution = Fraction(*page0.tags["YResolution"].value)
        except KeyError:
            pass
        else:
            md["tiff.ResolutionUnit"] = resolution_unit.name
            md["tiff.XResolution"] = float(x_resolution)
            md["tiff.YResolution"] = float(y_resolution)

            RESUNIT = tifffile.TIFF.RESUNIT
            scale = {
                RESUNIT.INCH: 25400.0,
                RESUNIT.CENTIMETER: 10000.0,
                RESUNIT.MILLIMETER: 1000.0,
                RESUNIT.MICROMETER: 1.0,
                RESUNIT.NONE: None,
            }.get(resolution_unit, None)
            if scale is not None:
                try:
                    mpp_x = scale / x_resolution
                    mpp_y = scale / y_resolution
                except ArithmeticError:
                    pass
                else:
                    md[PROPERTY_NAME_MPP_X] = mpp_x
                    md[PROPERTY_NAME_MPP_Y] = mpp_y
        return md

    def parse(self) -> dict[str, Any]:
        fmt = self.detect_format(self._tf)
        fmt = fmt.replace("-", "_")  # generic-tiff
        return getattr(self, f"parse_{fmt}")()  # type: ignore

    def parse_aperio(self) -> dict[str, Any]:
        """parse Aperio SVS properties"""
        md = self.new_metadata()

        # parse metadata from description
        page = self._tf.pages[0]
        desc = _check_page_description_encoding(page)
        md.update(_parse_metadata_aperio(desc))
        md["tiff.ImageDescription"] = desc

        # get series 0
        series0 = self._tf.series[0]
        md["tiffslide.series-index"] = 0  # svs standard

        # in case mpp wasn't available recover from tags
        if not _has_mpp(md):
            md.update(self.recover_mpp(series0))

        # collect level info
        md.update(self.collect_level_info(series0))
        return md

    def parse_leica(self) -> dict[str, Any]:
        """parse leica SCN properties"""
        md = self.new_metadata()

        # parse metadata from scn xml
        desc = self._tf.scn_metadata
        md["tiff.ImageDescription"] = desc

        # get all leica info
        md.update(_parse_metadata_leica(desc))

        # fill tile-width / tile-height
        idx = md["tiffslide.series-index"]
        for lvl, page in enumerate(self._tf.series[idx]):
            md[f"tiffslide.level[{lvl}].tile-width"] = page.tilewidth
            md[f"tiffslide.level[{lvl}].tile-height"] = page.tilelength

        return md

    def parse_ventana(self) -> dict[str, Any]:
        warn(
            "no special ventana-format metadata parsing implemented yet!",
            stacklevel=2,
        )
        return self.parse_generic_tiff()

    def parse_hamamatsu(self) -> dict[str, Any]:
        warn(
            "hamamatsu-format metadata parsing only partially implemented!",
            stacklevel=2,
        )
        md = self.parse_generic_tiff()

        # collect hamamatsu tags
        tags = self._tf.series[0][0].tags
        tag_map = {
            "65421": "hamamatsu.SourceLens",
            "65422": "hamamatsu.XOffsetFromSlideCentre",
            "65423": "hamamatsu.YOffsetFromSlideCentre",
            "Model": "hamamatsu.Model",
        }
        for tf_t, ts_t in tag_map.items():
            tag = tags.get(tf_t)
            if tag:
                md[ts_t] = tag.value

        md[PROPERTY_NAME_VENDOR] = "hamamatsu"
        if "hamamatsu.SourceLens" in md:
            md[PROPERTY_NAME_OBJECTIVE_POWER] = md["hamamatsu.SourceLens"]

        return md

    def parse_generic_tiff(self) -> dict[str, Any]:
        # todo: need to handle more supported formats in the future
        md = self.new_metadata()

        # store the description
        page = self._tf.pages[0]
        desc = _check_page_description_encoding(page)
        md["tiff.ImageDescription"] = desc

        md["tiffslide.series-index"] = 0  # use series 0
        series0 = self._tf.series[0]

        # in case mpp wasn't available recover from tags
        if not _has_mpp(md):
            md.update(self.recover_mpp(series0))

        # collect level info
        md.update(self.collect_level_info(series0))
        return md

    def parse_philips_tiff(self) -> dict[str, Any]:
        """parse Philips tiff properties"""
        md = self.parse_generic_tiff()
        if self._tf.philips_metadata is None:
            return md
        philips_metadata = ElementTree.fromstring(self._tf.philips_metadata)

        def get_first_attribute_with_name(
            root: ElementTree.Element, name: str
        ) -> str | None:
            return next(root.iterfind(f".//Attribute[@Name='{name}']")).text

        md[PROPERTY_NAME_VENDOR] = get_first_attribute_with_name(
            philips_metadata, "DICOM_MANUFACTURER"
        )
        pixel_spacing_attribute = get_first_attribute_with_name(
            philips_metadata, "DICOM_PIXEL_SPACING"
        )
        if pixel_spacing_attribute is not None:
            pixel_spacings = [
                float(element.strip('"')) * 1000
                for element in pixel_spacing_attribute.split(" ")
            ]
            mpp_y, mpp_x = pixel_spacings[0], pixel_spacings[1]
            md[PROPERTY_NAME_MPP_X] = mpp_x
            md[PROPERTY_NAME_MPP_Y] = mpp_y
        return md


def _parse_metadata_aperio(desc: str) -> dict[str, Any]:
    """Aperio SVS metadata"""
    aperio_meta = svs_description_metadata(desc)
    assert "Header" in aperio_meta

    md = {
        PROPERTY_NAME_COMMENT: desc,
        PROPERTY_NAME_VENDOR: "aperio",
        PROPERTY_NAME_QUICKHASH1: None,
        PROPERTY_NAME_BACKGROUND_COLOR: None,
        PROPERTY_NAME_OBJECTIVE_POWER: aperio_meta.get("AppMag", None),
        PROPERTY_NAME_MPP_X: aperio_meta.get("MPP", None),
        PROPERTY_NAME_MPP_Y: aperio_meta.get("MPP", None),
        PROPERTY_NAME_BOUNDS_X: None,
        PROPERTY_NAME_BOUNDS_Y: None,
        PROPERTY_NAME_BOUNDS_WIDTH: None,
        PROPERTY_NAME_BOUNDS_HEIGHT: None,
    }
    md.update({f"aperio.{k}": v for k, v in sorted(aperio_meta.items())})
    return md


def _parse_metadata_leica(image_description: str) -> dict[str, Any]:
    """return the leica SCN properties"""
    # todo: clean up. this is pretty convoluted
    md: dict[str, Any] = {PROPERTY_NAME_COMMENT: image_description}

    dct = _xml_to_dict(image_description)
    collection = dct["scn"]["collection"]

    slide_x_nm = int(collection["@sizeX"])
    slide_y_nm = int(collection["@sizeY"])

    first_non_macro_idx: int | None = None
    lvl_resolutions = defaultdict(list)
    series_offsets_nm = {}
    levels_start_x_nm: int | None = None
    levels_start_y_nm: int | None = None
    levels_end_x_nm: int | None = None
    levels_end_y_nm: int | None = None

    for idx, image in enumerate(collection["image"]):
        image_x_nm = int(image["view"]["@sizeX"])
        image_y_nm = int(image["view"]["@sizeY"])
        offset_x_nm = int(image["view"]["@offsetX"])
        offset_y_nm = int(image["view"]["@offsetY"])

        is_macro_image = (
            offset_x_nm == 0
            and offset_y_nm == 0
            and image_x_nm == slide_x_nm
            and image_y_nm == slide_y_nm
        )
        if is_macro_image:
            continue
        if levels_start_x_nm is None or levels_start_x_nm > offset_x_nm:
            levels_start_x_nm = offset_x_nm
        if levels_start_y_nm is None or levels_start_y_nm > offset_y_nm:
            levels_start_y_nm = offset_y_nm
        level_end_x_nm = offset_x_nm + image_x_nm
        level_end_y_nm = offset_y_nm + image_y_nm
        if levels_end_x_nm is None or levels_end_x_nm < level_end_x_nm:
            levels_end_x_nm = level_end_x_nm
        if levels_end_y_nm is None or levels_end_y_nm < level_end_y_nm:
            levels_end_y_nm = level_end_y_nm

        if first_non_macro_idx is None:
            first_non_macro_idx = idx

            _scan = image["scanSettings"]
            obj_pow = float(_scan["objectiveSettings"]["objective"])
            aperture = float(_scan["illuminationSettings"]["numericalAperture"])
            isource = str(_scan["illuminationSettings"]["illuminationSource"])

            md.update(
                {
                    PROPERTY_NAME_VENDOR: "leica",
                    PROPERTY_NAME_OBJECTIVE_POWER: obj_pow,
                    "leica.aperture": aperture,
                    "leica.creation-date": str(image["creationDate"]),
                    "leica.device-model": str(image["device"]["@model"]),
                    "leica.device-version": str(image["device"]["@version"]),
                    "leica.illumination-source": isource,
                }
            )

        for lvl, info in enumerate(image["pixels"]["dimension"]):
            resolution = image_x_nm / int(info["@sizeX"])
            # image_y_nm / int(info["@sizeY"])  <-- openslide just uses X
            lvl_resolutions[lvl].append(resolution)

        series_offsets_nm[idx] = offset_y_nm, offset_x_nm

    if not lvl_resolutions:
        raise ValueError("no non-macro images in file")

    assert levels_start_x_nm is not None
    assert levels_start_y_nm is not None
    assert levels_end_x_nm is not None
    assert levels_end_y_nm is not None

    resolutions0 = np.array(lvl_resolutions[0])
    # allow some threshold
    _r_avg = resolutions0.mean()
    if np.any(np.abs((resolutions0 - _r_avg) / _r_avg) > 0.02):
        raise ValueError(
            f"non-macro images vary too much in resolution: {lvl_resolutions[0]!r}"
        )

    nm_per_px = min(lvl_resolutions[0])
    mpp = nm_per_px / 1000.0
    md[PROPERTY_NAME_MPP_X] = mpp
    md[PROPERTY_NAME_MPP_Y] = mpp
    md[PROPERTY_NAME_BOUNDS_X] = int(levels_start_x_nm / nm_per_px)
    md[PROPERTY_NAME_BOUNDS_Y] = int(levels_start_y_nm / nm_per_px)
    md[PROPERTY_NAME_BOUNDS_WIDTH] = int(
        (levels_end_x_nm - levels_start_x_nm) / nm_per_px
    )
    md[PROPERTY_NAME_BOUNDS_HEIGHT] = int(
        (levels_end_y_nm - levels_start_y_nm) / nm_per_px
    )
    slide_x_px = math.ceil(slide_x_nm / nm_per_px)
    slide_y_px = math.ceil(slide_y_nm / nm_per_px)

    level_shapes = []
    located_series = defaultdict(list)
    for lvl, resolutions in sorted(lvl_resolutions.items()):
        lvl_nm_per_px = min(resolutions)

        for srs, offset_nm in series_offsets_nm.items():
            # implicitly assuming axes="YXS" ... (might be wrong?)
            offset_px = (
                int(offset_nm[0] / lvl_nm_per_px),
                int(offset_nm[1] / lvl_nm_per_px),
                0,
            )
            located_series[srs].append(offset_px)
        assert len(set(map(len, located_series.values()))) == 1

        lvl_size_x = math.ceil(slide_x_nm / lvl_nm_per_px)
        lvl_size_y = math.ceil(slide_y_nm / lvl_nm_per_px)
        md[f"tiffslide.level[{lvl}].height"] = lvl_size_y
        md[f"tiffslide.level[{lvl}].width"] = lvl_size_x
        md[f"tiffslide.level[{lvl}].downsample"] = (
            (slide_x_px / lvl_size_x) + (slide_y_px / lvl_size_y)
        ) / 2.0
        level_shapes.append((lvl_size_y, lvl_size_x, 3))

    md["tiffslide.series-index"] = first_non_macro_idx
    md["tiffslide.series-axes"] = "YXS"  # todo: verify
    md["tiffslide.series-composition"] = SeriesCompositionInfo(
        level_shapes=level_shapes,
        located_series=located_series,
    )

    return md


# --- helper functions --------------------------------------------------------


def _xml_to_dict(xml: str) -> dict[str, Any]:
    """helper function to convert xml string to a dictionary"""
    x = ElementTree.fromstring(xml)  # nosec B314

    def _to_dict(e):  # type: ignore
        tag = e.tag[e.tag.find("}") + 1 :]
        d = {f"@{k}": v for k, v in e.attrib.items()}
        for c in e:
            key, val = _to_dict(c).popitem()
            if key not in d:
                d[key] = val
            elif not isinstance(d[key], list):
                d[key] = [d[key], val]
            else:
                d[key].append(val)
        if e.text and e.text.strip():
            if d:
                d["#text"] = e.text
            else:
                d = e.text
        return {tag: d}

    return _to_dict(x)  # type: ignore


def _label_series_axes(axes: str) -> tuple[str, ...]:
    """helper to make series shapes more understandable"""
    return tuple(tifffile.TIFF.AXES_LABELS[c] for c in axes)


def _has_mpp(md: dict[str, Any]) -> bool:
    """check if metadata has MPP defined"""
    return md[PROPERTY_NAME_MPP_X] is not None and md[PROPERTY_NAME_MPP_Y] is not None


def _clip(x: int, min_: int, max_: int) -> int:
    """clip a value to a range"""
    return min(max(x, min_), max_)


def _get_filename(obj: Any) -> str:
    """return a filename or a placeholder"""
    if isinstance(obj, ReferenceFileSystem):
        for name in obj.templates.values():
            return name or ""
        else:
            return ""
    else:
        try:
            return obj.filename or ""
        except AttributeError:
            return ""


def _check_page_description_encoding(page: TiffPage) -> str:
    """try to return the description tag of a tiffpage"""
    value = page.description
    if isinstance(value, str) and value:
        return value

    # value was empty, try to recover
    value = page.tags.valueof(270, default="")

    if isinstance(value, str):
        return value

    elif isinstance(value, bytes) and value:
        raise TiffFileError(
            "ImageDescription tag has incompatible encoding.\n"
            "# Try setting the environment variable TIFFSLIDE_MONKEY_PATCH_NON_ASCII_DESCRIPTIONS=1\n"
            "# Or better fix your original file by running:\n"
            "$ python -m tiffslide.repair description-tag-encoding <WSI_FILENAME>",
        )

    else:
        raise ValueError(f"unsupported description type: {type(value)!r}")


# === monkey patching =================================================


def _env2bool(value: str) -> bool:
    """convert an environment variable to bool"""
    if not isinstance(value, str):
        raise TypeError(f"value must be of type str, got: {type(value).__name__}")
    value = value.lower()
    if value in {"yes", "true", "t", "y", "1"}:
        return True
    elif value in {"no", "false", "f", "n", "0", ""}:
        return False
    else:
        raise ValueError(f"could not parse value: {value!r}")


def _monkey_patch() -> None:
    """applying monkey tiffslide patches if requested"""
    if _env2bool(os.getenv("TIFFSLIDE_MONKEY_PATCH_NON_ASCII_DESCRIPTIONS", "")):
        from tiffslide.repair import monkey_patch_description_tag_encoding

        warn(
            "TIFFSLIDE_MONKEY_PATCH_NON_ASCII_DESCRIPTIONS is active!",
            RuntimeWarning,
            stacklevel=4,
        )
        monkey_patch_description_tag_encoding()



import json
import os
import pathlib
import subprocess
import sys

import matplotlib.pyplot as plt
import pandas as pd


def run_pytest_benchmarks(force: bool) -> None:
    td = os.environ["OPENSLIDE_TESTDATA_DIR"]
    assert td, f"must set OPENSLIDE_TESTDATA_DIR. Got: {td!r}"
    root = pathlib.Path(__file__).parent.parent
    assert root.joinpath("tiffslide", "tests").is_dir()

    fn_json = ".tiffslide_docs_benchmark.json"

    if not root.joinpath(fn_json).is_file() or force:
        subprocess.run(
            ["pytest", "-x", f"--benchmark-json={fn_json}", "--benchmark-only"],
            env=os.environ,
            cwd=root,
            check=True,
        )

    with root.joinpath(fn_json).open() as f:
        data = json.load(f)

    records = []
    for record in data["benchmarks"]:
        test_name, _, _ = record["name"].partition("[")
        test_name = test_name[5:]
        for rtime in record["stats"]["data"]:
            ps = record["params"].copy()
            file_type, modname = ps.pop("slide_with_tile_size")
            records.append(
                {
                    "test_name": test_name,
                    "file_type": file_type,
                    "modname": modname,
                    "label": "-".join(ps.values()),
                    "time": rtime,
                }
            )

    df = pd.DataFrame.from_records(records)
    for test_name, ptdf in df.groupby("test_name"):
        ft_ax_order = [
            "svs",
            "svs_jp2k",
            "generic",
            "hamamatsu",
            "leica",
            "ventana",
        ]
        assert set(ptdf["file_type"].unique()) == set(ft_ax_order)
        fig, axes = plt.subplots(1, len(ft_ax_order), figsize=(10, 4))
        fig.suptitle(test_name, x=0.1, y=0.99)
        ft_ax_map = {ft: ax for ft, ax in zip(ft_ax_order, axes)}

        for file_type, tdf in ptdf.groupby("file_type"):
            ldf = tdf.groupby(["modname", "label"])[["time"]].mean()
            ldf["time"] *= 1000
            ldf = ldf.reset_index()
            pdf = ldf.pivot(index="label", columns="modname", values="time")
            idx = sorted(pdf.index.tolist())
            pdf = pdf.loc[idx, :]
            pdf.plot.barh(title=f"{file_type}", ax=ft_ax_map[file_type], legend=False)

        axes[0].set_ylabel("access pattern")
        for ax in axes[1:]:
            ax.set_ylabel("")
            ax.set_yticklabels([])
        for ax in axes:
            ax.set_xlabel("time (ms)")
        handles, labels = axes[0].get_legend_handles_labels()
        fig.legend(handles[::-1], labels[::-1], loc="upper right")

        fig.savefig(root.joinpath("docs", "images", f"benchmark_{test_name}.png"))


if __name__ == "__main__":
    force = sys.argv[-1] == "--force"
    run_pytest_benchmarks(force)



from __future__ import annotations

import os
import sys
import warnings
from typing import IO
from typing import TYPE_CHECKING
from typing import Any
from typing import Protocol
from typing import TypedDict
from typing import Union
from typing import runtime_checkable

if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias

if TYPE_CHECKING:
    from types import TracebackType

    from fsspec import AbstractFileSystem

    if sys.version_info >= (3, 10):
        from types import EllipsisType
    else:
        EllipsisType: TypeAlias = Any


__all__ = [
    "PathLikeStr",
    "PathOrFileOrBufferLike",
    "OpenFileLike",
    "TiffFileIO",
    "Point3D",
    "Size3D",
    "Slice3D",
    "SeriesCompositionInfo",
]

if TYPE_CHECKING:
    from typing import AnyStr

    PathLikeAnyStr: TypeAlias = os.PathLike[AnyStr]


def __getattr__(name: str) -> Any:
    if name == "PathLikeAnyStr":
        warnings.warn(
            "use PathLikeStr instead of PathLikeAnyStr",
            DeprecationWarning,
            stacklevel=2,
        )
        return "os.PathLike[AnyStr]"
    else:
        raise AttributeError(name)


PathLikeStr: TypeAlias = "os.PathLike[str]"


@runtime_checkable
class OpenFileLike(Protocol):
    """minimal fsspec open file type"""

    fs: AbstractFileSystem
    path: str

    def __enter__(self) -> IO[bytes]:
        ...

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        ...


@runtime_checkable
class TiffFileIO(Protocol):
    """minimal stream io for use in tifffile.FileHandle"""

    def seek(self, offset: int, whence: int = 0) -> int:
        ...

    def tell(self) -> int:
        ...

    def read(self, n: int = -1) -> bytes:
        ...

    def readinto(self, __buffer: Any) -> int | None:
        ...


PathOrFileOrBufferLike = Union[str, PathLikeStr, OpenFileLike, TiffFileIO]


Point3D: TypeAlias = "tuple[int, int, int]"
Size3D: TypeAlias = "tuple[int, int, int]"
Slice2D: TypeAlias = "tuple[slice, slice]"
_Slice2D_YX0: TypeAlias = "tuple[slice, slice, EllipsisType]"
Slice3D: TypeAlias = "tuple[slice, slice, slice] | _Slice2D_YX0"


class SeriesCompositionInfo(TypedDict):
    """composition information for combining tifffile series

    Notes
    -----
    level_shapes:
        A list of 3D sizes for each composited level
    located_series:
        A dict mapping the TiffFile series index to a list of
        3D offsets for each level for compositing

    """

    level_shapes: list[Size3D]
    located_series: dict[int, list[Point3D]]



"""
provides helpers for handling and compositing arrays and zarr-like groups
"""
from __future__ import annotations

import json
from typing import TYPE_CHECKING
from typing import Any
from typing import Iterator
from typing import Mapping

import numpy as np
import zarr
from fsspec.implementations.reference import ReferenceFileSystem
from tifffile import TiffFile
from tifffile import ZarrTiffStore
from zarr.storage import BaseStore
from zarr.storage import FSStore
from zarr.storage import KVStore as _KVStore

from tiffslide._compat import NotTiffFile
from tiffslide._types import Point3D
from tiffslide._types import SeriesCompositionInfo
from tiffslide._types import Size3D
from tiffslide._types import Slice3D

if TYPE_CHECKING:
    from numpy.typing import DTypeLike
    from numpy.typing import NDArray


__all__ = [
    "get_zarr_store",
    "get_zarr_depth_and_dtype",
    "get_zarr_selection",
]


# --- zarr storage classes --------------------------------------------

if "__contains__" not in _KVStore.__dict__:
    # fix missing contains caused double decode:
    # https://github.com/bayer-science-for-a-better-life/tiffslide/issues/72#issuecomment-1627918238
    class KVStore(_KVStore):
        def __contains__(self, item: str) -> bool:
            return item in self._mutable_mapping

else:
    KVStore = _KVStore  # type: ignore


class _CompositedStore(Mapping[str, Any]):
    """prefix zarr stores to allow mounting them in groups"""

    def __init__(
        self,
        prefixed_stores: Mapping[str, Mapping[str, Any]],
        *,
        zattrs: Mapping[str, Any] | None = None,
    ) -> None:
        grp = zarr.group({})
        if zattrs:
            grp.attrs["tiffslide.series-composition"] = zattrs
        self._base = grp.store
        self._stores = {}
        for prefix, store in prefixed_stores.items():
            assert not prefix.endswith("/")
            self._stores[prefix] = store

    def __len__(self) -> int:
        return sum(map(len, self._stores.values())) + len(self._base)

    def __contains__(self, item: object) -> bool:
        if not isinstance(item, str):
            return False

        prefix, _, key = item.partition("/")
        if key:
            try:
                store = self._stores[prefix]
            except KeyError:
                pass
            else:
                return key in store

        return item in self._base or prefix in self._stores

    def __iter__(self) -> Iterator[str]:
        for prefix, store in self._stores.items():
            yield from (f"{prefix}/{key}" for key in store.keys())
        yield from self._base.keys()

    def __getitem__(self, item: str) -> Any:
        prefix, _, key = item.partition("/")

        if key:
            try:
                store = self._stores[prefix]
            except KeyError:
                pass
            else:
                return store[key]

        return self._base[item]


def _get_series_zarr(
    obj: TiffFile | ReferenceFileSystem,
    series_idx: int,
    *,
    num_decode_threads: int | None = None,
) -> Mapping[str, Any]:
    """return a zarr store from the object"""
    if isinstance(obj, (TiffFile, NotTiffFile)):
        zstore = obj.series[series_idx].aszarr(maxworkers=num_decode_threads)  # type: ignore
    elif isinstance(obj, ReferenceFileSystem):
        zstore = FSStore(f"s{series_idx}", fs=obj)
    else:
        raise NotImplementedError(f"{type(obj).__name__} unsupported")
    return zstore  # type: ignore


def get_zarr_store(
    properties: Mapping[str, Any],
    tf: TiffFile | ReferenceFileSystem | None,
    *,
    num_decode_threads: int | None = None,
) -> BaseStore:
    """return a zarr store

    Parameters
    ----------
    properties:
        the TiffSlide().properties mapping
    tf:
        the corresponding TiffFile instance
    num_decode_threads:
        number of threads used for decoding (default num_cpu / 2)

    Returns
    -------
    store:
        a zarr store of the tiff
    """
    if tf is None:
        raise NotImplementedError("support in future versions")

    # the tiff might contain multiple series that require composition
    composition: SeriesCompositionInfo | None = properties.get(
        "tiffslide.series-composition"
    )
    store: BaseStore
    if composition:
        prefixed_stores = {}
        for series_idx in composition["located_series"].keys():
            _store = _get_series_zarr(
                tf, series_idx, num_decode_threads=num_decode_threads
            )
            # encapsulate store as group if tifffile returns a zarr array
            if ".zarray" in _store:
                _store = _CompositedStore({"0": _store})
            prefixed_stores[str(series_idx)] = _store

        _store = _CompositedStore(prefixed_stores, zattrs=composition)
        store = KVStore(_store)

    else:
        series_idx = properties.get("tiffslide.series-index", 0)
        _store = _get_series_zarr(tf, series_idx, num_decode_threads=num_decode_threads)

        # encapsulate store as group if tifffile returns a zarr array
        if ".zarray" in _store:
            _store = _CompositedStore({"0": _store})
            store = KVStore(_store)
        elif isinstance(_store, BaseStore):
            store = _store
        else:
            store = KVStore(_store)

    return store


def get_zarr_selection(
    grp: zarr.Group,
    level: int,
    selection: Slice3D,
) -> NDArray[np.int_]:
    """retrieve the selection of the zarr Group"""
    composition: SeriesCompositionInfo = grp.attrs.get("tiffslide.series-composition")

    if composition is None:
        # no composition required, simply retrieve the array
        return grp[str(level)][selection]

    else:
        # we need to composite the array
        # todo: getting these overlaps should be done via an RTree...
        overlaps = {}
        dtype = None
        fill_value = None
        level_shape = composition["level_shapes"][level]
        located_series = composition["located_series"]

        for series_idx, level_offsets in located_series.items():
            arr = grp[f"{series_idx}/{level}"]
            offset = level_offsets[level]

            if dtype is None:
                dtype = arr.dtype
            if fill_value is None:
                fill_value = arr.fill_value

            overlap = get_overlap(selection, level_shape, offset, arr.shape)
            if overlap is None:
                continue

            # todo: could take a shortcut here...
            # target, source = overlap
            # if target == item:  # < this check is incorrect [:, :] != [0:10, 0:10]
            #     return arr[source]

            overlaps[series_idx] = overlap

        s0, s1, s2 = selection
        r0 = range(level_shape[0])[s0]
        r1 = range(level_shape[1])[s1]
        r2 = range(level_shape[2])[s2 if s2 is not ... else slice(None)]
        shape = (r0.stop - r0.start, r1.stop - r1.start, r2.stop - r2.start)
        # todo: optimize! In the most common case this should be filled entirely
        out = np.full(shape, fill_value=fill_value, dtype=dtype)

        for series_idx, (target, source) in overlaps.items():
            arr = grp[f"{series_idx}/{level}"]
            out[target] = arr[source]

    return out


def get_zarr_chunk_sizes(
    grp: zarr.Group,
    *,
    level: int = 0,
    sum_axis: int | None = None,
) -> NDArray[np.int64]:
    """return an array of the raw chunk byte sizes

    EXPERIMENTAL --- do not rely on this

    """
    store = grp.store

    if not isinstance(store, ZarrTiffStore):
        while hasattr(store, "_mutable_mapping"):
            # noinspection PyProtectedMember
            store = store._mutable_mapping

        # noinspection PyProtectedMember
        if isinstance(store, _CompositedStore) and {"0"} == set(store._stores):
            # noinspection PyProtectedMember
            store = store._stores["0"]

        while hasattr(store, "_mutable_mapping"):
            # noinspection PyProtectedMember
            store = store._mutable_mapping

        if not isinstance(store, ZarrTiffStore):
            raise NotImplementedError(f"store type: {type(store).__name__!r}")

    for key, value in store.items():
        if ".zarray" in key:
            levelstr = (key.split("/")[0] + "/") if "/" in key else ""
            # skip if not selected level
            if levelstr == "" and level != 0:
                continue
            elif levelstr == "" and level == 0:
                break
            elif levelstr != f"{level}/":
                continue
            else:
                break
    else:
        raise ValueError(f"no matching level: {level}")

    value = json.loads(value)
    shape = value["shape"]
    chunks = value["chunks"]

    assert len(shape) == len(chunks)
    if len(shape) not in (2, 3):
        raise NotImplementedError("chunk dimensions not in (2, 3)")

    chunked = tuple(i // j + (1 if i % j else 0) for i, j in zip(shape, chunks))

    # fixme:
    #  relies on private functionality of ZarrTiffStore, might break at any time
    try:
        # noinspection PyProtectedMember
        chunkmode = store._chunkmode
        # noinspection PyProtectedMember
        parse_key = store._parse_key
    except AttributeError:
        raise RuntimeError("probably not supported with your tifffile version")

    chunk_sizes: NDArray[np.int64] = np.full(chunked, dtype=np.int64, fill_value=-1)

    # _index = ""
    for indices in np.ndindex(*chunked):
        chunkindex = ".".join(str(index) for index in indices)
        key = levelstr + chunkindex
        keyframe, page, _, offset, bytecount = parse_key(key)
        # key = levelstr + _index + chunkindex
        if page and chunkmode and offset is None:
            # offset = page.dataoffsets[0]
            bytecount = keyframe.nbytes
        if offset and bytecount:
            chunk_sizes[indices] = bytecount

    if chunk_sizes.ndim not in (2, 3):
        raise NotImplementedError("chunk dimensions not in (2, 3)")

    if sum_axis is None:
        return chunk_sizes  # type: ignore
    else:
        return chunk_sizes.sum(axis=sum_axis)  # type: ignore


# --- helper functions ------------------------------------------------


def get_overlap(
    selection: Slice3D,
    cshape: Size3D,
    offset: Point3D,
    ashape: Size3D,
) -> tuple[Slice3D, Slice3D] | None:
    """return correct slices for overlapping arrays

    CompositedArray
    +------------------------+
    | x------------x         |
    | | Slice o----|---o     |
    | |       |    |   |     |
    | x------------x   |     |
    |         | Array  |     |
    |         o--------o     |
    +------------------------+

    Notes
    -----
    if selection is a slice out of the CompositedArray with shape=cshape,
    this function returns either None, if the slice and the Array share
    no overlap, or two slices (target and source), so that:

    `new_arr[target] = array[source]`

    can be used to assign the overlap correctly to the output array.

    """
    s0, s1, s2 = selection
    d0, d1, d2 = cshape
    o0, o1, o2 = offset
    a0, a1, a2 = ashape

    # dim 0
    x0, x1, _ = s0.indices(d0)
    X0 = o0
    X1 = o0 + a0
    if X1 < x0 or x1 <= X0:
        return None
    # dim 1
    y0, y1, _ = s1.indices(d1)
    Y0 = o1
    Y1 = o1 + a1
    if Y1 < y0 or y1 <= Y0:
        return None
    # dim 2
    if s2 is Ellipsis:
        z0, z1 = 0, 1
    else:
        z0, z1, _ = s2.indices(d2)
    Z0 = o2
    Z1 = o2 + a2
    if Z1 < z0 or z1 <= Z0:
        return None

    ix0 = max(x0, X0)
    iy0 = max(y0, Y0)
    iz0 = max(z0, Z0)
    ix1 = min(x1, X1)
    iy1 = min(y1, Y1)
    iz1 = min(z1, Z1)

    target = (
        slice(ix0 - x0, ix1 - x0, 1),
        slice(iy0 - y0, iy1 - y0, 1),
        slice(iz0 - z0, iz1 - z0, 1),
    )
    source = (
        slice(ix0 - X0, ix1 - X0, 1),
        slice(iy0 - Y0, iy1 - Y0, 1),
        slice(iz0 - Z0, iz1 - Z0, 1),
    )
    return target, source


def verify_located_arrays(
    shape: Size3D, located_arrays: dict[Point3D, zarr.Array]
) -> tuple[str, Any]:
    """verify located arrays

    ensures that dtypes and fill_values agree and that
    arrays are within the composited boundaries.
    """
    dt_fv = set()
    for loc, arr in located_arrays.items():
        dt_fv.add((arr.dtype, arr.fill_value))

        x0, x1, x2 = loc
        d0, d1, d2 = arr.shape
        out_of_bounds = (
            (x0 + d0 < 0 or x0 >= shape[0])
            or (x1 + d1 < 0 or x1 >= shape[1])
            or (x2 + d2 < 0 or x2 >= shape[2])
        )
        if out_of_bounds:
            raise ValueError(f"array at location {loc!r} is out of bounds")

    if not len(dt_fv) == 1:
        raise ValueError(
            f"arrays don't share the same dtype and fill_value, got: {dt_fv!r}"
        )
    ((dtype, fill_value),) = dt_fv
    return dtype, fill_value


def get_zarr_depth_and_dtype(grp: zarr.Group, axes: str) -> tuple[int, DTypeLike]:
    """return the image depth from the zarr group"""
    if "tiffslide.series-composition" in grp.attrs:
        srs = next(iter(grp.attrs["series-composition"]["located_series"]))
        key = f"{srs}/0"
    else:
        key = "0"  # -> level

    zarray: zarr.core.Array = grp[key]

    if axes == "YXS":
        depth = zarray.shape[2]
    elif axes == "CYX":
        depth = zarray.shape[0]
    elif axes == "YX":
        depth = 1
    else:
        raise NotImplementedError(f"axes={axes!r}")

    return depth, zarray.dtype
